
from __future__ import annotations

from collections.abc import Awaitable, Callable, Iterable
from typing import Optional, Union, NoReturn

from pymap.bytes import MaybeBytes
from pymap.concurrent import Event
from pymap.config import IMAPConfig
from pymap.context import socket_info, connection_exit
from pymap.exceptions import NotAllowedError, NotSupportedError, \
    CloseConnection
from pymap.fetch import MessageAttributes
from pymap.interfaces.login import LoginInterface
from pymap.interfaces.session import SessionInterface
from pymap.parsing.command import CommandAuth, CommandNonAuth, CommandSelect, \
    Command
from pymap.parsing.command.any import CapabilityCommand, LogoutCommand, \
    NoOpCommand
from pymap.parsing.command.nonauth import AuthenticateCommand, LoginCommand, \
    StartTLSCommand
from pymap.parsing.command.auth import AppendCommand, CreateCommand, \
    DeleteCommand, ListCommand, RenameCommand, SelectCommand, StatusCommand, \
    SubscribeCommand, UnsubscribeCommand
from pymap.parsing.command.select import CheckCommand, CloseCommand, \
    IdleCommand, ExpungeCommand, CopyCommand, MoveCommand, FetchCommand, \
    StoreCommand, SearchCommand
from pymap.parsing.commands import InvalidCommand
from pymap.parsing.primitives import List, Number
from pymap.parsing.response import ResponseOk, ResponseNo, ResponseBad, \
    ResponseCode, ResponsePreAuth, CommandResponse, UntaggedResponse
from pymap.parsing.response.code import Capability, PermanentFlags, UidNext, \
    UidValidity, Unseen, MailboxId
from pymap.parsing.response.specials import FlagsResponse, ExistsResponse, \
    RecentResponse, FetchResponse, ListResponse, LSubResponse, \
    SearchResponse, StatusResponse
from pymap.parsing.specials import StatusAttribute, FetchAttribute, FetchValue
from pymap.selected import SelectedMailbox
from pysasl import AuthenticationCredentials

__all__ = ['ConnectionState']

_AuthCommands = Union[AuthenticateCommand, LoginCommand]
_CommandRet = tuple[CommandResponse, Optional[SelectedMailbox]]
_CommandFunc = Callable[[Command], Awaitable[_CommandRet]]

_flags_attr = FetchAttribute(b'FLAGS')
_uid_attr = FetchAttribute(b'UID')


class ConnectionState:
    """Defines the state flow of the IMAP connection. Determines if a command
    can be processed at that point in the connection, and interacts with the
    backend plugin.

    """

    def __init__(self, login: LoginInterface, config: IMAPConfig) -> None:
        super().__init__()
        self.config = config
        self.auth = config.initial_auth
        self.login = login
        self._session: Optional[SessionInterface] = None
        self._selected: Optional[SelectedMailbox] = None
        self._capability = list(config.initial_capability)

    @property
    def session(self) -> SessionInterface:
        if self._session is None:
            raise RuntimeError()  # State checking should prevent this.
        return self._session

    @property
    def selected(self) -> SelectedMailbox:
        if self._selected is None:
            raise RuntimeError()  # State checking should prevent this.
        return self._selected

    @property
    def capability(self) -> Capability:
        if self._session:
            return Capability(self._capability)
        else:
            if self.auth.get_server(b'PLAIN') is None:
                logindisabled = [b'LOGINDISABLED']
            else:
                logindisabled = []
            return Capability(self._capability + logindisabled +
                              [b'AUTH=%b' % mech.name for mech in
                               self.auth.server_mechanisms])

    async def do_cleanup(self) -> None:
        try:
            await self.session.cleanup()
        except Exception:
            pass

    async def _login(self, creds: AuthenticationCredentials) \
            -> SessionInterface:
        stack = connection_exit.get()
        identity = await self.login.authenticate(creds)
        return await stack.enter_async_context(identity.new_session())

    async def do_greeting(self) -> CommandResponse:
        sock_info = socket_info.get()
        if self.config.reject_dnsbl and sock_info.dnsbl is not None:
            raise NotAllowedError(f'Connection rejected: {sock_info.dnsbl}')
        preauth_creds = self.config.preauth_credentials
        if preauth_creds:
            self._session = await self._login(preauth_creds)
        elif sock_info.from_localhost:
            self.auth = self.config.tls_auth
        resp_cls = ResponsePreAuth if preauth_creds else ResponseOk
        return resp_cls(b'*', self.config.greeting, self.capability)

    async def do_authenticate(self, cmd: _AuthCommands,
                              creds: Optional[AuthenticationCredentials]) \
            -> CommandResponse:
        if not creds:
            return ResponseNo(cmd.tag, b'Invalid authentication mechanism.')
        self._session = await self._login(creds)
        self._capability.extend(self.config.login_capability)
        return ResponseOk(cmd.tag, b'Authentication successful.',
                          self.capability)

    async def do_login(self, cmd: LoginCommand) -> _CommandRet:
        if b'LOGINDISABLED' in self.capability:
            raise NotSupportedError('LOGIN is disabled.')
        creds = AuthenticationCredentials(
            cmd.userid.decode('utf-8', 'surrogateescape'),
            cmd.password.decode('utf-8', 'surrogateescape'))
        return await self.do_authenticate(cmd, creds), None

    async def do_starttls(self, cmd: StartTLSCommand) -> _CommandRet:
        try:
            self._capability.remove(b'STARTTLS')
        except ValueError:
            raise NotSupportedError('STARTTLS not available.')
        self.auth = self.config.tls_auth
        return ResponseOk(cmd.tag, b'Ready to handshake.'), None

    async def do_capability(self, cmd: CapabilityCommand) -> _CommandRet:
        response = ResponseOk(cmd.tag, b'Capabilities listed.')
        response.add_untagged(UntaggedResponse(self.capability.string))
        return response, None

    async def do_noop(self, cmd: NoOpCommand) -> _CommandRet:
        updates = None
        if self._selected and self._session:
            updates = await self.session.check_mailbox(self.selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_select(self, cmd: SelectCommand) -> _CommandRet:
        self._selected = None
        mailbox, updates = await self.session.select_mailbox(
            cmd.mailbox, cmd.readonly)
        if updates.readonly:
            num_recent = mailbox.recent
            resp = ResponseOk(cmd.tag, b'Selected mailbox.',
                              ResponseCode.of(b'READ-ONLY'))
            resp.add_untagged_ok(b'Read-only mailbox.', PermanentFlags([]))
        else:
            num_recent = updates.session_flags.recent
            resp = ResponseOk(cmd.tag, b'Selected mailbox.',
                              ResponseCode.of(b'READ-WRITE'))
            resp.add_untagged_ok(b'Flags permitted.',
                                 PermanentFlags(mailbox.permanent_flags))
        messages = updates.messages
        resp.add_untagged(FlagsResponse(mailbox.flags))
        resp.add_untagged(ExistsResponse(messages.exists))
        resp.add_untagged(RecentResponse(num_recent))
        resp.add_untagged_ok(b'Predicted next UID.',
                             UidNext(mailbox.next_uid))
        resp.add_untagged_ok(b'UIDs valid.',
                             UidValidity(mailbox.uid_validity))
        if mailbox.first_unseen:
            resp.add_untagged_ok(b'First unseen message.',
                                 Unseen(mailbox.first_unseen))
        resp.add_untagged_ok(b'Object ID.', MailboxId(mailbox.mailbox_id))
        return resp, updates

    async def do_create(self, cmd: CreateCommand) -> _CommandRet:
        if cmd.mailbox == 'INBOX':
            return ResponseNo(cmd.tag, b'Cannot create INBOX.'), None
        mailbox_id, updates = await self.session.create_mailbox(
            cmd.mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.',
                          MailboxId(mailbox_id)), updates

    async def do_delete(self, cmd: DeleteCommand) -> _CommandRet:
        if cmd.mailbox == 'INBOX':
            return ResponseNo(cmd.tag, b'Cannot delete INBOX.'), None
        updates = await self.session.delete_mailbox(
            cmd.mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_rename(self, cmd: RenameCommand) -> _CommandRet:
        if cmd.to_mailbox == 'INBOX':
            return ResponseNo(cmd.tag, b'Cannot rename to INBOX.'), None
        updates = await self.session.rename_mailbox(
            cmd.from_mailbox, cmd.to_mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_status(self, cmd: StatusCommand) -> _CommandRet:
        mailbox, updates = await self.session.get_mailbox(
            cmd.mailbox, selected=self._selected)
        data: dict[StatusAttribute, MaybeBytes] = {}
        for attr in cmd.status_list:
            if attr == b'MESSAGES':
                data[attr] = Number(mailbox.exists)
            elif attr == b'RECENT':
                if updates and updates.mailbox_id == mailbox.mailbox_id:
                    data[attr] = Number(updates.session_flags.recent)
                else:
                    data[attr] = Number(mailbox.recent)
            elif attr == b'UNSEEN':
                data[attr] = Number(mailbox.unseen)
            elif attr == b'UIDNEXT':
                data[attr] = Number(mailbox.next_uid)
            elif attr == b'UIDVALIDITY':
                data[attr] = Number(mailbox.uid_validity)
            elif attr == b'MAILBOXID':
                data[attr] = mailbox.mailbox_id.parens
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        resp.add_untagged(StatusResponse(cmd.mailbox, data))
        return resp, updates

    async def do_append(self, cmd: AppendCommand) -> _CommandRet:
        if len(cmd.messages) > 1 and b'MULTIAPPEND' not in self.capability:
            raise NotSupportedError('MULTIAPPEND is disabled.')
        if cmd.cancelled:
            return ResponseNo(cmd.tag, b'APPEND cancelled.'), None
        if cmd.error:
            raise cmd.error
        append_uid, updates = await self.session.append_messages(
            cmd.mailbox, cmd.messages, selected=self._selected)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.', append_uid)
        return resp, updates

    async def do_subscribe(self, cmd: SubscribeCommand) -> _CommandRet:
        updates = await self.session.subscribe(
            cmd.mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_unsubscribe(self, cmd: UnsubscribeCommand) -> _CommandRet:
        updates = await self.session.unsubscribe(
            cmd.mailbox, selected=self._selected)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_list(self, cmd: ListCommand) -> _CommandRet:
        mailboxes, updates = await self.session.list_mailboxes(
            cmd.ref_name, cmd.filter, subscribed=cmd.only_subscribed,
            selected=self._selected)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        resp_type = LSubResponse if cmd.only_subscribed else ListResponse
        for name, sep, attrs in mailboxes:
            resp.add_untagged(resp_type(name, sep, attrs))
        return resp, updates

    async def do_check(self, cmd: CheckCommand) -> _CommandRet:
        updates = await self.session.check_mailbox(
            self.selected, housekeeping=True)
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), updates

    async def do_close(self, cmd: CloseCommand) -> _CommandRet:
        await self.session.expunge_mailbox(self.selected)
        self._selected = None
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), None

    async def do_expunge(self, cmd: ExpungeCommand) -> _CommandRet:
        updates = await self.session.expunge_mailbox(
            self.selected, cmd.uid_set)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        return resp, updates

    async def do_copy(self, cmd: CopyCommand) -> _CommandRet:
        copy_uid, updates = await self.session.copy_messages(
            self.selected, cmd.sequence_set, cmd.mailbox)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.', copy_uid)
        return resp, updates

    async def do_move(self, cmd: MoveCommand) -> _CommandRet:
        copy_uid, updates = await self.session.move_messages(
            self.selected, cmd.sequence_set, cmd.mailbox)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        resp.add_untagged_ok(b'Moved.', copy_uid)
        return resp, updates

    async def do_fetch(self, cmd: FetchCommand) -> _CommandRet:
        if not cmd.uid:
            self.selected.hide_expunged = True
        set_seen = not self.selected.readonly and \
            any(attr.set_seen for attr in cmd.attributes)
        messages, updates = await self.session.fetch_messages(
            self.selected, cmd.sequence_set, set_seen)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        for msg_seq, msg in messages:
            if msg.expunged:
                resp.code = ResponseCode.of(b'EXPUNGEISSUED')
            msg_attrs = MessageAttributes(msg, self.selected, cmd.attributes)
            fetch_resp = FetchResponse(msg_seq, msg_attrs,
                                       writing_hook=msg_attrs.load_hook())
            resp.add_untagged(fetch_resp)
        return resp, updates

    async def do_search(self, cmd: SearchCommand) -> _CommandRet:
        if not cmd.uid:
            self.selected.hide_expunged = True
        messages, updates = await self.session.search_mailbox(
            self.selected, cmd.keys)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        msg_ids: list[int] = []
        for msg_seq, msg in messages:
            if msg.expunged:
                resp.code = ResponseCode.of(b'EXPUNGEISSUED')
            if cmd.uid:
                msg_ids.append(msg.uid)
            else:
                msg_ids.append(msg_seq)
        resp.add_untagged(SearchResponse(msg_ids))
        return resp, updates

    async def do_store(self, cmd: StoreCommand) -> _CommandRet:
        if not cmd.uid:
            self.selected.hide_expunged = True
        if cmd.silent:
            self.selected.silence(cmd.sequence_set, cmd.flag_set, cmd.mode)
        messages, updates = await self.session.update_flags(
            self.selected, cmd.sequence_set, cmd.flag_set, cmd.mode)
        resp = ResponseOk(cmd.tag, cmd.command + b' completed.')
        session_flags = self.selected.session_flags
        for msg_seq, msg in messages:
            if msg.expunged:
                resp.code = ResponseCode.of(b'EXPUNGEISSUED')
            elif cmd.silent:
                continue
            flags = msg.get_flags(session_flags)
            fetch_data: list[FetchValue] = [
                FetchValue.of(_flags_attr, List(flags, sort=True))]
            if cmd.uid:
                fetch_data.append(
                    FetchValue.of(_uid_attr, Number(msg.uid)))
            resp.add_untagged(FetchResponse(msg_seq, fetch_data))
        return resp, updates

    async def do_idle(self, cmd: IdleCommand) -> _CommandRet:
        if b'IDLE' not in self.capability:
            raise NotSupportedError('IDLE is disabled.')
        return ResponseOk(cmd.tag, cmd.command + b' completed.'), None

    @classmethod
    async def do_logout(cls, cmd: LogoutCommand) -> NoReturn:
        raise CloseConnection()

    async def receive_updates(self, cmd: IdleCommand, done: Event) \
            -> Iterable[UntaggedResponse]:
        selected = await self.session.check_mailbox(
            self.selected, wait_on=done)
        self._selected, untagged = selected.fork(cmd)
        return untagged

    @classmethod
    def _get_func_name(cls, cmd: Command) -> str:
        cmd_type = type(cmd)
        while cmd_type.delegate:
            cmd_type = cmd_type.delegate
        cmd_str = str(cmd_type.command, 'ascii').lower()
        return 'do_' + cmd_str

    async def do_command(self, cmd: Command) -> CommandResponse:
        if isinstance(cmd, InvalidCommand):
            return ResponseBad(cmd.tag, cmd.message)
        elif self._session and isinstance(cmd, CommandNonAuth):
            msg = cmd.command + b': Already authenticated.'
            return ResponseBad(cmd.tag, msg)
        elif not self._session and isinstance(cmd, CommandAuth):
            msg = cmd.command + b': Must authenticate first.'
            return ResponseBad(cmd.tag, msg)
        elif not self._selected and isinstance(cmd, CommandSelect):
            msg = cmd.command + b': Must select a mailbox first.'
            return ResponseBad(cmd.tag, msg)
        func_name = self._get_func_name(cmd)
        try:
            func: _CommandFunc = getattr(self, func_name)
        except AttributeError:
            return ResponseNo(cmd.tag, cmd.command + b': Not Implemented')
        response, selected = await func(cmd)
        if selected is not None:
            self._selected, untagged = selected.fork(cmd)
            response.add_untagged(*untagged)
        return response




from __future__ import annotations

from collections.abc import Iterable, Iterator, Sequence
from itertools import chain
from typing import Union, Optional

from .. import Params, Parseable, Space
from ..exceptions import NotParseable
from ...bytes import rev

__all__ = ['MaxValue', 'SequenceSet']

_SeqIdx = Union['MaxValue', int]
_SeqElem = Union[_SeqIdx, tuple[_SeqIdx, _SeqIdx]]


class MaxValue:
    """The type used as a placeholder for the maximum value."""

    def __eq__(self, other) -> bool:
        if isinstance(other, MaxValue):
            return True
        return NotImplemented

    def __hash__(self) -> int:
        return hash(type(self))


class SequenceSet(Parseable[Sequence[_SeqElem]]):
    """Represents a sequence set from an IMAP stream.

    Args:
        sequences: The sequence set data.
        uid: True if the sequences refer to message UIDs.

    Attributes:
        uid: True if the sequences refer to message UIDs.

    """

    _num_pattern = rev.compile(br'[1-9]\d*')
    _max = MaxValue()

    def __init__(self, sequences: Sequence[_SeqElem],
                 uid: bool = False) -> None:
        super().__init__()
        self.sequences = sequences
        self.uid = uid
        self._raw: Optional[bytes] = None

    @classmethod
    def all(cls, uid: bool = False) -> SequenceSet:
        """A sequence set intended to contain all values."""
        return _AllSequenceSet(uid)

    @property
    def value(self) -> Sequence[_SeqElem]:
        """The sequence set data."""
        return self.sequences

    @property
    def is_all(self) -> bool:
        """True if the sequence set starts at ``1`` and ends at the maximum
        value.

        This may be used to optimize cases of checking for a value in the set,
        avoiding the need to provide ``max_value`` in :meth:`.flatten` or
        :meth:`.iter`.

        """
        first = self.sequences[0]
        return isinstance(first, tuple) \
            and first[0] == 1 and isinstance(first[1], MaxValue)

    @classmethod
    def _get_range(cls, elem: _SeqElem, max_value: int) -> Iterable[int]:
        if isinstance(elem, int):
            if elem <= max_value:
                return range(elem, elem + 1)
            else:
                return ()
        elif isinstance(elem, MaxValue):
            return range(max_value, max_value + 1)
        else:
            left, right = elem
            if isinstance(left, MaxValue):
                left = max_value
            if isinstance(right, MaxValue):
                right = max_value
            low = min(left, right)
            if low <= max_value:
                high = min(max(left, right), max_value)
                return range(low, high + 1)
            else:
                return ()

    def flatten(self, max_value: int) -> frozenset[int]:
        """Return a set of all values contained in the sequence set.

        Args:
            max_value: The maximum value, in place of any ``*``.

        """
        return frozenset(self.iter(max_value))

    def iter(self, max_value: int) -> Iterator[int]:
        """Iterates through the sequence numbers contained in the set, bounded
        by the given maximum value (in place of any ``*``).

        Args:
            max_value: The maximum value of the set.

        """
        return chain.from_iterable(
            (self._get_range(elem, max_value) for elem in self.sequences))

    def _elem_bytes(self, elem: _SeqIdx) -> bytes:
        if isinstance(elem, MaxValue):
            return b'*'
        else:
            return b'%d' % elem

    def __bytes__(self) -> bytes:
        if self._raw is not None:
            return self._raw
        parts = []
        for group in self.value:
            if isinstance(group, tuple):
                left = self._elem_bytes(group[0])
                right = self._elem_bytes(group[1])
                parts.append(b'%b:%b' % (left, right))
            else:
                parts.append(self._elem_bytes(group))
        self._raw = raw = b','.join(parts)
        return raw

    def __eq__(self, other) -> bool:
        if isinstance(other, SequenceSet):
            return self.uid == other.uid \
                and self.sequences == other.sequences
        return super().__eq__(other)

    def __hash__(self) -> int:
        return hash((type(self), tuple(self.sequences), self.uid))

    def __repr__(self) -> str:
        attr = 'uidset' if self.uid else 'set'
        return '<SequenceSet %s=%r>' % (attr, self.sequences)

    @classmethod
    def _parse_part(cls, buf: memoryview) -> tuple[_SeqElem, memoryview]:
        if buf and buf[0] == 0x2a:
            item1: _SeqIdx = cls._max
            buf = buf[1:]
        else:
            match = cls._num_pattern.match(buf)
            if match:
                buf = buf[match.end(0):]
                item1 = int(match.group(0))
            else:
                raise NotParseable(buf)
        if buf and buf[0] == 0x3a:
            buf = buf[1:]
            if buf and buf[0] == 0x2a:
                return (item1, cls._max), buf[1:]
            match = cls._num_pattern.match(buf)
            if match:
                buf = buf[match.end(0):]
                return (item1, int(match.group(0))), buf
            raise NotParseable(buf)
        return item1, buf

    @classmethod
    def build(cls, seqs: Iterable[int], uid: bool = False) -> SequenceSet:
        """Build a new sequence set that contains the given values using as
        few groups as possible.

        Args:
            seqs: The sequence values to build.
            uid: True if the sequences refer to message UIDs.

        """
        seqs_list = sorted(set(seqs))
        groups: list[Union[int, tuple[int, int]]] = []
        group: Union[int, tuple[int, int]] = seqs_list[0]
        for i in range(1, len(seqs_list)):
            group_i = seqs_list[i]
            if isinstance(group, int):
                if group_i == group + 1:
                    group = (group, group_i)
                else:
                    groups.append(group)
                    group = group_i
            elif isinstance(group, tuple):
                if group_i == group[1] + 1:
                    group = (group[0], group_i)
                else:
                    groups.append(group)
                    group = group_i
        groups.append(group)
        return SequenceSet(groups, uid)

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[SequenceSet, memoryview]:
        try:
            _, buf = Space.parse(buf, params)
        except NotParseable:
            pass
        sequences = []
        while buf:
            item, buf = cls._parse_part(buf)
            sequences.append(item)
            if buf and buf[0] != 0x2c:
                break
            buf = buf[1:]
        if not sequences:
            raise NotParseable(buf)
        return cls(sequences, uid=params.uid), buf


class _AllSequenceSet(SequenceSet):

    def __init__(self, uid: bool) -> None:
        super().__init__([(1, self._max)], uid)

    def iter(self, max_value: int) -> Iterator[int]:
        return iter(range(1, max_value + 1))

    def __bytes__(self) -> bytes:
        return b'1:*'

    def __repr__(self) -> str:
        return '<SequenceSet set=all>'




from __future__ import annotations

from collections.abc import Sequence
from typing import Optional

from aioredis import Redis, ResponseError

from .keys import NamespaceKeys, FilterKeys
from .scripts.filter import FilterScripts
from pymap.filter import PluginFilterSet
from pymap.parsing.specials import ObjectId

__all__ = ['FilterSet']

_scripts = FilterScripts()


class FilterSet(PluginFilterSet[bytes]):

    _active_name = b''

    def __init__(self, redis: Redis, ns_keys: NamespaceKeys) -> None:
        super().__init__(None, bytes)
        self._redis = redis
        self._keys = FilterKeys(ns_keys)

    async def put(self, name: str, value: bytes) -> None:
        keys = self._keys
        name_bytes = name.encode('utf-8')
        key = ObjectId.random(b'S').value
        pipe = self._redis.pipeline()
        pipe.hset(keys.data, key, value)
        pipe.hset(keys.names, name_bytes, key)
        await pipe.execute()

    async def delete(self, name: str) -> None:
        try:
            await _scripts.delete(self._redis, self._keys,
                                  name=name.encode('utf-8'),
                                  active_name=self._active_name)
        except ResponseError as exc:
            if 'filter not found' in str(exc):
                raise KeyError(name) from exc
            elif 'filter is active' in str(exc):
                raise ValueError(name) from exc
            raise

    async def rename(self, before_name: str, after_name: str) -> None:
        try:
            await _scripts.rename(self._redis, self._keys,
                                  before_name=before_name.encode('utf-8'),
                                  after_name=after_name.encode('utf-8'))
        except ResponseError as exc:
            if 'filter not found' in str(exc):
                raise KeyError(before_name) from exc
            elif 'filter already exists' in str(exc):
                raise KeyError(after_name) from exc
            raise

    async def clear_active(self) -> None:
        keys = self._keys
        await self._redis.hdel(keys.names, self._active_name)

    async def set_active(self, name: str) -> None:
        try:
            await _scripts.set_active(self._redis, self._keys,
                                      name=name.encode('utf-8'),
                                      active_name=self._active_name)
        except ResponseError as exc:
            if 'filter not found' in str(exc):
                raise KeyError(name) from exc
            raise

    async def get(self, name: str) -> bytes:
        try:
            return await _scripts.get(self._redis, self._keys,
                                      name=name.encode('utf-8'))
        except ResponseError as exc:
            if 'filter not found' in str(exc):
                raise KeyError(name) from exc
            raise

    async def get_active(self) -> Optional[bytes]:
        try:
            return await _scripts.get(self._redis, self._keys,
                                      name=self._active_name)
        except ResponseError as exc:
            if 'filter not found' in str(exc):
                return None
            raise

    async def get_all(self) -> tuple[Optional[str], Sequence[str]]:
        keys = self._keys
        filter_names = await self._redis.hgetall(keys.names)
        active_key: Optional[bytes] = filter_names.get(self._active_name)
        active_name: Optional[str] = None
        names: list[str] = []
        for name, key in filter_names.items():
            if name != self._active_name:
                name_str = name.decode('utf-8')
                names.append(name_str)
                if key == active_key:
                    active_name = name_str
        return active_name, names




from __future__ import annotations

from typing import Optional

from pymap.exceptions import AppendFailure
from pymap.interfaces.filter import FilterInterface, FilterCompilerInterface
from pymap.parsing.message import AppendMessage
from sievelib.parser import Parser

from .runner import SieveRunner
from .util import unquote

__all__ = ['SieveParseError', 'SieveCompiler', 'SieveFilter']


class SieveParseError(ValueError):
    """Parsing the sieve script failed."""
    pass


class SieveCompiler(FilterCompilerInterface[bytes]):
    """Compiles sieve scripts into :class:`SieveFilter` objects."""

    #: The list of Sieve extensions supported by the compiler.
    extensions: list[bytes] = [
        b'fileinto', b'reject', b'envelope', b'body']

    def __init__(self) -> None:
        super().__init__()
        self.parser = Parser()

    @property
    def value_type(self) -> type[bytes]:
        return bytes

    @property
    def filter_type(self) -> type[SieveFilter]:
        return SieveFilter

    async def compile(self, value: bytes) -> SieveFilter:
        parser = self.parser
        try:
            success = parser.parse(value)
        except Exception as exc:
            raise SieveParseError('Unhandled parsing error') from exc
        if success:
            runner = SieveRunner(parser.result)
            return SieveFilter(runner)
        else:
            raise SieveParseError(parser.error)


class SieveFilter(FilterInterface):
    """Filter implementation for applying sieve scripts to the appended
    message. The sieve script applies various conditions and controls to
    ultimately produce a set of actions to apply to the message.

    Args:
        runner: The sieve script runner.

    """

    def __init__(self, runner: SieveRunner) -> None:
        super().__init__()
        self.runner = runner

    async def apply(self, sender: str, recipient: str, mailbox: str,
                    append_msg: AppendMessage) \
            -> tuple[Optional[str], AppendMessage]:
        for action in self.runner.get_actions(sender, recipient, append_msg):
            if action.name == 'keep':
                pass
            elif action.name == 'fileinto':
                mailbox = unquote(action.arguments['mailbox'])
                return mailbox, append_msg
            elif action.name == 'reject':
                msg = unquote(action.arguments['text']).encode('utf-8')
                raise AppendFailure(mailbox, msg)
            elif action.name == 'discard':
                return None, append_msg
            else:
                msg = b'Unrecognized sieve action: %b' \
                    % action.name.encode('ascii')
                raise AppendFailure(mailbox, msg) from KeyError(action.name)
        return mailbox, append_msg




from __future__ import annotations

import asyncio
import sys
from argparse import ArgumentParser, SUPPRESS
from asyncio import Task, CancelledError
from collections.abc import Sequence
from contextlib import AsyncExitStack
from datetime import datetime, timedelta, timezone
from ssl import SSLContext
from typing import Optional

from grpclib.events import listen, RecvRequest
from grpclib.health.check import ServiceStatus
from grpclib.health.service import Health, OVERALL
from grpclib.server import Server
from pymap.context import cluster_metadata
from pymap.interfaces.backend import ServiceInterface
from pymapadmin import is_compatible, __version__ as server_version
from pymapadmin.local import socket_file, token_file

from .errors import get_incompatible_version_error
from .handlers import handlers
from .typing import Handler

__all__ = ['AdminService']


class AdminService(ServiceInterface):  # pragma: no cover
    """A pymap service implemented using a `grpc <https://grpc.io/>`_ server
    to perform admin functions on a running backend.

    This service can be exposed in *private* or *public* mode. Private mode
    does not use SSL/TLS encryption and authentication is not required, so it
    should be appropriately firewalled. Public mode requires both SSL/TLS
    encryption and the client provide credentials for the user being
    operated on, e.g. a user can change their own password but cannot change
    another user's password.

    The `pymap-admin <https://github.com/icgood/pymap-admin>`_ CLI may be used
    to interact with the admin service and serves as a reference client
    implementation.

    """

    @classmethod
    def add_arguments(cls, parser: ArgumentParser) -> None:
        group = parser.add_argument_group('admin service')
        group.add_argument('--admin-path', action=socket_file.add_action,
                           help=SUPPRESS)
        group.add_argument('--admin-host', action='append', metavar='IFACE',
                           help='the network interface to listen on')
        group.add_argument('--admin-port', metavar='NUM', type=int,
                           default=50051, help='the port to listen on')
        group.add_argument('--admin-token-file', action=token_file.add_action,
                           help=SUPPRESS)
        group.add_argument('--admin-token-duration', metavar='SEC', type=int,
                           help=SUPPRESS)

    def _init_admin_token(self) -> None:
        expiration: Optional[datetime] = None
        if self.config.args.admin_token_duration:
            duration_sec: int = self.config.args.admin_token_duration
            duration = timedelta(seconds=duration_sec)
            expiration = datetime.now(tz=timezone.utc) + duration
        admin_key = self.config.admin_key
        token = self.backend.login.tokens.get_admin_token(
            admin_key, expiration=expiration)
        if token is not None:
            self._write_admin_token(token)
            print(f'PYMAP_ADMIN_TOKEN={token}', file=sys.stderr)
        if admin_key is not None:
            cluster_metadata.get().local['admin'] = admin_key

    def _write_admin_token(self, admin_token: str) -> None:
        try:
            path = token_file.get_temp(mkdir=True)
            path.write_text(admin_token)
        except OSError:
            pass

    async def _check_version(self, event: RecvRequest) -> None:
        client_version = event.metadata.get('client-version')
        if client_version is not None:
            if isinstance(client_version, bytes):
                client_version = client_version.decode('ascii')
            if not is_compatible(client_version, server_version):
                raise get_incompatible_version_error(
                    client_version, server_version)

    def _get_health(self) -> Handler:
        backend_status = ServiceStatus()
        self.backend.status.register(backend_status.set)
        return Health({OVERALL: [backend_status]})

    async def start(self, stack: AsyncExitStack) -> None:
        self._init_admin_token()
        backend = self.backend
        path: Optional[str] = self.config.args.admin_path
        host: Optional[str] = self.config.args.admin_host
        port: Optional[int] = self.config.args.admin_port
        server_handlers: list[Handler] = [self._get_health()]
        server_handlers.extend(handler(backend) for _, handler in handlers)
        ssl = self.config.ssl_context
        local_task = await self._start_local(server_handlers, path)
        remote_task = await self._start(server_handlers, host, port, ssl)
        stack.callback(local_task.cancel)
        stack.callback(remote_task.cancel)

    def _new_server(self, server_handlers: Sequence[Handler]) -> Server:
        server = Server(server_handlers)
        listen(server, RecvRequest, self._check_version)
        return server

    async def _start_local(self, server_handlers: Sequence[Handler],
                           path: Optional[str]) -> Task[None]:
        server = self._new_server(server_handlers)
        path = str(socket_file.get_temp(mkdir=True))
        await server.start(path=path)
        return asyncio.create_task(self._run(server))

    async def _start(self, server_handlers: Sequence[Handler],
                     host: Optional[str], port: Optional[int],
                     ssl: Optional[SSLContext]) -> Task[None]:
        server = self._new_server(server_handlers)
        await server.start(host=host, port=port, ssl=ssl, reuse_address=True)
        return asyncio.create_task(self._run(server))

    async def _run(self, server: Server) -> None:
        try:
            await server.wait_closed()
        except CancelledError:
            server.close()
            await server.wait_closed()
            raise




from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional

from ..parsing.specials import Flag, ExtensionOptions

__all__ = ['AppendMessage']


@dataclass(frozen=True)
class AppendMessage:
    """A single message from the APPEND command.

    Args:
        literal: The message literal.
        when: The internal timestamp to assign to the message.
        flag_set: The flags to assign to the message.
        options: The extension options in use for the message.

    """

    literal: bytes
    when: Optional[datetime] = None
    flag_set: frozenset[Flag] = field(default_factory=frozenset)
    options: ExtensionOptions = field(default_factory=ExtensionOptions.empty)




from __future__ import annotations

from abc import abstractmethod, ABCMeta
from collections.abc import Sequence
from typing import TypeVar, Generic, Final, Optional

__all__ = ['ParsingExpectedT', 'ParsingState', 'ParsingInterrupt',
           'ParsingExpectation']

#: Type variable for expected data during parsing.
ParsingExpectedT = TypeVar('ParsingExpectedT')


class ParsingState:
    """Contains mutable parsing state. As parsing advances, it may need
    external data to proceed. Because parsing is linear, these expectations
    are defined as sequence of objects that are iterated during parsing.

    Warning:
        The expectation keys and object types are dynamic, to prevent cyclic
        dependencies in the parsing system.

    See Also:
        :class:`pymap.parsing.interrupts.ParsingInterrupt`

    Args:
        continuations: The IMAP literal string continuations.

    """

    __slots__ = ['continuations']

    def __init__(self, *, continuations: Sequence[memoryview] = None) -> None:
        super().__init__()
        self.continuations: Final = iter(continuations or ())


class ParsingInterrupt(Exception):
    """Exception thrown to interrupt the IMAP parsing process, because an
    expectation was not met. The caller should update provide an updated
    :class:`ParsingState` object that meets the expectation.

    Args:
        expected: The required data expectation.

    """

    __slots__ = ['expected']

    def __init__(self, expected: ParsingExpectation) -> None:
        super().__init__()
        self.expected: Final = expected


class ParsingExpectation(Generic[ParsingExpectedT], metaclass=ABCMeta):
    """Base class for parsing expectations that may require additional data
    before parsing can advance.

    """

    __slots__: Sequence[str] = []

    @abstractmethod
    def consume(self, state: ParsingState) -> Optional[ParsingExpectedT]:
        """Consume and return a piece of data, if available.

        Args:
            state: The parsing state.

        """
        ...

    def expect(self, state: ParsingState) -> ParsingExpectedT:
        """Indicates that the buffer has been successfully parsed so far, but
        additional data is expected from the parsing state.

        Args:
            state: The parsing state object.

        Raises:
            ParsingInterrupt: The expected data was not available.

        """
        next_val = self.consume(state)
        if next_val is None:
            raise ParsingInterrupt(self)
        else:
            return next_val


class ExpectContinuation(ParsingExpectation[memoryview]):
    """Indicates that the buffer has been successfully parsed so far, but
    requires a continuation of the command from the client.

    Args:
        message: The message from the server.
        literal_length: If the continuation is for a string literal, this
            is the byte length to expect.

    """

    __slots__ = ['message', 'literal_length']

    def __init__(self, message: bytes, literal_length: int = 0) -> None:
        super().__init__()
        self.message: Final = message
        self.literal_length: Final = literal_length

    def consume(self, state: ParsingState) -> Optional[memoryview]:
        return next(state.continuations, None)



"""Utilities for managing IMAP searches."""

from __future__ import annotations

import re
from abc import abstractmethod, ABCMeta
from collections.abc import Iterable
from datetime import datetime
from typing import AnyStr, Optional, Final

from .exceptions import SearchNotAllowed
from .interfaces.message import MessageInterface, LoadedMessageInterface
from .parsing.specials import ObjectId, SearchKey, SequenceSet
from .parsing.specials.flag import Flag, Answered, Deleted, Draft, Flagged, \
    Recent, Seen
from .selected import SelectedMailbox

__all__ = ['SearchParams', 'SearchCriteria', 'SearchCriteriaSet']


class SearchParams:
    """Defines certain parameters and routines necessary to process any
    kind of search criteria. If a parameter is not supplied, or a method not
    implemented, any search keys that require it will fail.

    Args:
        selected: The active mailbox session.
        max_seq: The highest message sequence ID in the mailbox.
        max_uid: The highest message UID in the mailbox.
        disabled: Search keys that should be disabled.

    """

    __slots__ = ['selected', 'disabled', 'max_seq', 'max_uid', 'session_flags']

    def __init__(self, selected: SelectedMailbox, *,
                 disabled: Iterable[bytes] = None) -> None:
        self.selected: Final = selected
        self.disabled: Final = frozenset(disabled or [])
        self.max_seq: Final = selected.messages.exists
        self.max_uid: Final = selected.messages.max_uid
        self.session_flags: Final = selected.session_flags


class SearchCriteria(metaclass=ABCMeta):
    """Base class for different types of search criteria.

    Args:
        params: The parameters that may be used by some searches.

    """

    def __init__(self, params: SearchParams) -> None:
        self.params = params

    @classmethod
    def _in(cls, substr: AnyStr, data: AnyStr) -> bool:
        re_flags = re.I | re.A
        escaped_substr = re.escape(substr)
        return re.search(escaped_substr, data, re_flags) is not None

    @abstractmethod
    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        """Implemented by sub-classes to define the search criteria.

        Args:
            msg_seq: The message sequence ID.
            msg: The message object.
            loaded_msg: The message content.

        """
        ...

    @classmethod
    def of(cls, key: SearchKey, params: SearchParams) -> SearchCriteria:
        """Factory method for producing a search criteria sub-class from a
        search key.

        Args:
            key: The search key defining the criteria.
            params: The parameters that may be used by some searches.

        """
        key_name = key.value
        if key_name in params.disabled:
            raise SearchNotAllowed(key)
        elif key.inverse:
            return InverseSearchCriteria(key.not_inverse, params)
        elif key_name == b'SEQSET':
            return SequenceSetSearchCriteria(key.filter_sequence_set, params)
        elif key_name == b'KEYSET':
            return SearchCriteriaSet(key.filter_key_set, params)
        elif key_name == b'ALL':
            return AllSearchCriteria(params)
        elif key_name == b'OR':
            left_key, right_key = key.filter_key_or
            return OrSearchCriteria(left_key, right_key, params)
        elif key_name == b'EMAILID':
            return HasEmailIdSearchCriteria(key.filter_object_id, params)
        elif key_name == b'THREADID':
            return HasThreadIdSearchCriteria(key.filter_object_id, params)
        elif key_name == b'ANSWERED':
            return HasFlagSearchCriteria(Answered, True, params)
        elif key_name == b'UNANSWERED':
            return HasFlagSearchCriteria(Answered, False, params)
        elif key_name == b'DELETED':
            return HasFlagSearchCriteria(Deleted, True, params)
        elif key_name == b'UNDELETED':
            return HasFlagSearchCriteria(Deleted, False, params)
        elif key_name == b'DRAFT':
            return HasFlagSearchCriteria(Draft, True, params)
        elif key_name == b'UNDRAFT':
            return HasFlagSearchCriteria(Draft, False, params)
        elif key_name == b'FLAGGED':
            return HasFlagSearchCriteria(Flagged, True, params)
        elif key_name == b'UNFLAGGED':
            return HasFlagSearchCriteria(Flagged, False, params)
        elif key_name == b'RECENT':
            return HasFlagSearchCriteria(Recent, True, params)
        elif key_name == b'OLD':
            return HasFlagSearchCriteria(Recent, False, params)
        elif key_name == b'SEEN':
            return HasFlagSearchCriteria(Seen, True, params)
        elif key_name == b'UNSEEN':
            return HasFlagSearchCriteria(Seen, False, params)
        elif key_name == b'KEYWORD':
            return HasFlagSearchCriteria(key.filter_flag, True, params)
        elif key_name == b'UNKEYWORD':
            return HasFlagSearchCriteria(key.filter_flag, False, params)
        elif key_name == b'NEW':
            return NewSearchCriteria(params)
        elif key_name == b'BEFORE':
            return DateSearchCriteria(key.filter_datetime, '<', params)
        elif key_name == b'ON':
            return DateSearchCriteria(key.filter_datetime, '=', params)
        elif key_name == b'SINCE':
            return DateSearchCriteria(key.filter_datetime, '>=', params)
        elif key_name == b'SENTBEFORE':
            return HeaderDateSearchCriteria(key.filter_datetime, '<', params)
        elif key_name == b'SENTON':
            return HeaderDateSearchCriteria(key.filter_datetime, '=', params)
        elif key_name == b'SENTSINCE':
            return HeaderDateSearchCriteria(key.filter_datetime, '>=', params)
        elif key_name == b'SMALLER':
            return SizeSearchCriteria(key.filter_int, '<', params)
        elif key_name == b'LARGER':
            return SizeSearchCriteria(key.filter_int, '>', params)
        elif key_name in (b'BCC', b'CC', b'FROM', b'SUBJECT', b'TO'):
            return EnvelopeSearchCriteria(key_name, key.filter_str, params)
        elif key_name == b'HEADER':
            name, value = key.filter_header
            return HeaderSearchCriteria(name, value, params)
        elif key_name in (b'BODY', b'TEXT'):
            return BodySearchCriteria(key.filter_str, params)
        raise SearchNotAllowed(key)


class SearchCriteriaSet(SearchCriteria):
    """Search criteria composed of a set of search criteria that must all
    match. If the set is empty, nothing will match.

    Args:
        keys: The set of search keys that must match.
        params: The parameters that may be used by some searches.

    """

    def __init__(self, keys: frozenset[SearchKey],
                 params: SearchParams) -> None:
        super().__init__(params)
        self.all_criteria = [SearchCriteria.of(key, params) for key in keys]

    @property
    def sequence_set(self) -> SequenceSet:
        """The sequence set to use when finding the messages to match against.
        This will default to all messages unless the search criteria set
        contains a sequence set.

        """
        try:
            seqset_crit = next(crit for crit in self.all_criteria
                               if isinstance(crit, SequenceSetSearchCriteria))
        except StopIteration:
            return SequenceSet.all()
        else:
            return seqset_crit.seq_set

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        """The message matches if all the defined search key criteria match.

        Args:
            msg_seq: The message sequence ID.
            msg: The message object.

        """
        return all(crit.matches(msg_seq, msg, loaded_msg)
                   for crit in self.all_criteria)


class InverseSearchCriteria(SearchCriteria):
    """Matches only if the given search criteria does not match."""

    def __init__(self, key: SearchKey, params: SearchParams) -> None:
        super().__init__(params)
        self.key = SearchCriteria.of(key, params)

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        return not self.key.matches(msg_seq, msg, loaded_msg)


class AllSearchCriteria(SearchCriteria):
    """Always matches anything."""

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        return True


class OrSearchCriteria(SearchCriteria):
    """Matches if either of the search criteria match."""

    def __init__(self, left: SearchKey, right: SearchKey,
                 params: SearchParams) -> None:
        super().__init__(params)
        self.left = SearchCriteria.of(left, self.params)
        self.right = SearchCriteria.of(right, self.params)

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        return (self.left.matches(msg_seq, msg, loaded_msg)
                or self.right.matches(msg_seq, msg, loaded_msg))


class SequenceSetSearchCriteria(SearchCriteria):
    """Matches if the message is contained in the sequence set."""

    def __init__(self, seq_set: SequenceSet, params: SearchParams) -> None:
        super().__init__(params)
        self.seq_set = seq_set
        if seq_set.uid:
            self.flat = seq_set.flatten(params.max_uid)
        else:
            self.flat = seq_set.flatten(params.max_seq)

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        if self.seq_set.uid:
            return msg.uid in self.flat
        else:
            return msg_seq in self.flat


class HasEmailIdSearchCriteria(SearchCriteria):
    """Matches if the message has the same
    :attr:`~pymap.interfaces.message.MessageInterface.email_id` value.

    """

    def __init__(self, email_id: ObjectId, params: SearchParams) -> None:
        super().__init__(params)
        self.email_id = email_id

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        try:
            email_id = msg.email_id
        except ValueError:
            return False
        else:
            return self.email_id == email_id


class HasThreadIdSearchCriteria(SearchCriteria):
    """Matches if the message has the same
    :attr:`~pymap.interfaces.message.MessageInterface.thread_id` value.

    """

    def __init__(self, thread_id: ObjectId, params: SearchParams) -> None:
        super().__init__(params)
        self.thread_id = thread_id

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        return self.thread_id == msg.thread_id


class HasFlagSearchCriteria(SearchCriteria):
    """Matches if the message has the given flag in their permanent or
    session flag sets.

    """

    def __init__(self, flag: Flag, expected: bool,
                 params: SearchParams) -> None:
        super().__init__(params)
        self.flag = flag
        self.expected = expected

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        has_flag = self.flag in msg.get_flags(self.params.session_flags)
        expected = self.expected
        return (has_flag and expected) or (not expected and not has_flag)


class NewSearchCriteria(SearchCriteria):
    """Matches if the message is considered "new", i.e. recent and unseen."""

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        flags = msg.get_flags(self.params.session_flags)
        return Recent in flags and Seen not in flags


class DateSearchCriteria(SearchCriteria):
    """Matches by comparing against the internal date of the message."""

    def __init__(self, when: datetime, op: str, params: SearchParams) -> None:
        super().__init__(params)
        self.when = when.date()
        self.op = op

    @classmethod
    def _get_msg_date(cls, msg: MessageInterface,
                      loaded_msg: LoadedMessageInterface) \
            -> Optional[datetime]:
        return msg.internal_date

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        msg_datetime = self._get_msg_date(msg, loaded_msg)
        if msg_datetime is None:
            return False
        msg_date = msg_datetime.date()
        if self.op == '<':  # BEFORE
            return msg_date < self.when
        elif self.op == '=':  # ON
            return msg_date == self.when
        elif self.op == '>=':  # SINCE
            return msg_date >= self.when
        raise ValueError(self.op)


class HeaderDateSearchCriteria(DateSearchCriteria):
    """Matches by comparing against the ``Date:`` header of the message."""

    @classmethod
    def _get_msg_date(cls, msg: MessageInterface,
                      loaded_msg: LoadedMessageInterface) \
            -> Optional[datetime]:
        envelope = loaded_msg.get_envelope_structure()
        return envelope.date.datetime if envelope.date else None


class SizeSearchCriteria(SearchCriteria):
    """Matches by comparing against the size of the message."""

    def __init__(self, size: int, op: str, params: SearchParams) -> None:
        super().__init__(params)
        self.size = size
        self.op = op

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        size = loaded_msg.get_size()
        if self.op == '<':
            return size < self.size
        elif self.op == '>':
            return size > self.size
        raise ValueError(self.op)


class EnvelopeSearchCriteria(SearchCriteria):
    """Matches by checking for strings within various fields of the envelope
    structure.

    """

    def __init__(self, key: bytes, value: str, params: SearchParams) -> None:
        super().__init__(params)
        self.key = key
        self.value = value

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        envelope = loaded_msg.get_envelope_structure()
        if self.key == b'BCC':
            if not envelope.bcc:
                return False
            return any(self._in(self.value, str(bcc)) for bcc in envelope.bcc)
        elif self.key == b'CC':
            if not envelope.cc:
                return False
            return any(self._in(self.value, str(cc)) for cc in envelope.cc)
        elif self.key == b'FROM':
            if not envelope.from_:
                return False
            return any(self._in(self.value, str(from_))
                       for from_ in envelope.from_)
        elif self.key == b'SUBJECT':
            if not envelope.subject:
                return False
            return self._in(self.value, str(envelope.subject))
        elif self.key == b'TO':
            if not envelope.to:
                return False
            return any(self._in(self.value, str(to)) for to in envelope.to)
        raise ValueError(self.key)


class HeaderSearchCriteria(SearchCriteria):
    """Matches if the message has a header containing a value."""

    def __init__(self, name: str, value: str, params: SearchParams) -> None:
        super().__init__(params)
        self.name = name.encode('ascii')
        self.value = value

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        values = loaded_msg.get_header(self.name)
        return any(self._in(self.value, value) for value in values)


class BodySearchCriteria(SearchCriteria):
    """Matches if the message body contains a value."""

    def __init__(self, value: str, params: SearchParams) -> None:
        super().__init__(params)
        self.value = bytes(value, 'utf-8', 'replace')

    def matches(self, msg_seq: int, msg: MessageInterface,
                loaded_msg: LoadedMessageInterface) -> bool:
        return loaded_msg.contains(self.value)




from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import ClassVar, Final, Optional

from pymap.parsing import Parseable, Params, EndLine
from pymap.parsing.exceptions import NotParseable
from pymap.parsing.primitives import Atom, Number, String, QuotedString

__all__ = ['Command', 'NoOpCommand', 'CapabilityCommand', 'StartTLSCommand',
           'AuthenticateCommand', 'UnauthenticateCommand', 'LogoutCommand',
           'HaveSpaceCommand', 'PutScriptCommand', 'ListScriptsCommand',
           'SetActiveCommand', 'GetScriptCommand', 'DeleteScriptCommand',
           'RenameScriptCommand', 'CheckScriptCommand']


class Command(Parseable[bytes]):
    """Base class for ManageSieve commands."""

    #: The command key, e.g. ``b'NOOP'``.
    command: ClassVar[bytes] = b''

    _commands: Optional[Mapping[bytes, type[Command]]] = None

    @property
    def value(self) -> bytes:
        """The command name."""
        return self.command

    def __bytes__(self) -> bytes:
        return self.command

    def __repr__(self) -> str:
        return '<%s>' % (type(self).__name__)

    @classmethod
    def _all_commands(cls) -> Sequence[type[Command]]:
        return [
            NoOpCommand, CapabilityCommand, StartTLSCommand,
            AuthenticateCommand, UnauthenticateCommand, LogoutCommand,
            HaveSpaceCommand, PutScriptCommand, ListScriptsCommand,
            SetActiveCommand, GetScriptCommand, DeleteScriptCommand,
            RenameScriptCommand, CheckScriptCommand]

    @classmethod
    def _get_commands(cls) -> Mapping[bytes, type[Command]]:
        if cls._commands is None:
            mapping = {}
            for cmd in cls._all_commands():
                mapping[cmd.command] = cmd
            cls._commands = mapping
        return cls._commands

    @classmethod
    def _parse_script_name(self, buf: memoryview, params: Params,
                           allow_empty: bool = False) \
            -> tuple[str, memoryview]:
        name_obj, after = String.parse(buf, params)
        if not allow_empty and not name_obj.value:
            raise NotParseable(buf)
        try:
            return name_obj.value.decode('utf-8'), after
        except UnicodeError as exc:
            raise NotParseable(buf) from exc

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[Command, memoryview]:
        commands = cls._get_commands()
        cmd_name, after = Atom.parse(buf, params)
        try:
            cmd_type = commands[cmd_name.value.upper()]
        except KeyError:
            raise NotParseable(buf)
        else:
            buf = after
        cmd, buf = cmd_type.parse(buf, params)
        return cmd, buf


class NoOpCommand(Command):

    command = b'NOOP'

    def __init__(self, tag: Optional[bytes]) -> None:
        super().__init__()
        self.tag: Final = tag

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[NoOpCommand, memoryview]:
        whitespace = cls._whitespace_length(buf)
        tag: Optional[bytes] = None
        if whitespace > 0:
            buf = buf[whitespace:]
            try:
                tag_obj, buf = String.parse(buf, params)
            except NotParseable:
                pass
            else:
                tag = tag_obj.value
        _, buf = EndLine.parse(buf, params)
        return cls(tag), buf


class CapabilityCommand(Command):

    command = b'CAPABILITY'

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[CapabilityCommand, memoryview]:
        _, buf = EndLine.parse(buf, params)
        return cls(), buf


class StartTLSCommand(Command):

    command = b'STARTTLS'

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[StartTLSCommand, memoryview]:
        _, buf = EndLine.parse(buf, params)
        return cls(), buf


class AuthenticateCommand(Command):

    command = b'AUTHENTICATE'

    def __init__(self, mech_name: bytes, initial_data: bytes = None) -> None:
        super().__init__()
        self.mech_name: Final = mech_name
        self.initial_data: Final = initial_data

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[AuthenticateCommand, memoryview]:
        mech_name, buf = QuotedString.parse(buf, params)
        try:
            data_obj, buf = String.parse(buf, params)
        except NotParseable:
            initial_data: Optional[bytes] = None
        else:
            initial_data = data_obj.value
        return cls(mech_name.value, initial_data), buf


class UnauthenticateCommand(Command):

    command = b'UNAUTHENTICATE'

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[UnauthenticateCommand, memoryview]:
        _, buf = EndLine.parse(buf, params)
        return cls(), buf


class LogoutCommand(Command):

    command = b'LOGOUT'

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[LogoutCommand, memoryview]:
        _, buf = EndLine.parse(buf, params)
        return cls(), buf


class HaveSpaceCommand(Command):

    command = b'HAVESPACE'

    def __init__(self, script_name: str, size: int) -> None:
        super().__init__()
        self.script_name: Final = script_name
        self.size: Final = size

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[HaveSpaceCommand, memoryview]:
        script_name, after = cls._parse_script_name(buf, params)
        if not script_name:
            raise NotParseable(buf)
        size, buf = Number.parse(after, params)
        _, buf = EndLine.parse(buf, params)
        return cls(script_name, size.value), buf


class PutScriptCommand(Command):

    command = b'PUTSCRIPT'

    def __init__(self, script_name: str, script_data: bytes) -> None:
        super().__init__()
        self.script_name: Final = script_name
        self.script_data: Final = script_data

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[PutScriptCommand, memoryview]:
        script_name, buf = cls._parse_script_name(buf, params)
        script_data, buf = String.parse(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(script_name, script_data.value), buf


class ListScriptsCommand(Command):

    command = b'LISTSCRIPTS'

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[ListScriptsCommand, memoryview]:
        _, buf = EndLine.parse(buf, params)
        return cls(), buf


class SetActiveCommand(Command):

    command = b'SETACTIVE'

    def __init__(self, script_name: Optional[str]) -> None:
        super().__init__()
        self.script_name: Final = script_name

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[SetActiveCommand, memoryview]:
        script_name, buf = cls._parse_script_name(buf, params, True)
        _, buf = EndLine.parse(buf, params)
        return cls(script_name or None), buf


class GetScriptCommand(Command):

    command = b'GETSCRIPT'

    def __init__(self, script_name: str) -> None:
        super().__init__()
        self.script_name: Final = script_name

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[GetScriptCommand, memoryview]:
        script_name, buf = cls._parse_script_name(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(script_name), buf


class DeleteScriptCommand(Command):

    command = b'DELETESCRIPT'

    def __init__(self, script_name: str) -> None:
        super().__init__()
        self.script_name: Final = script_name

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[DeleteScriptCommand, memoryview]:
        script_name, buf = cls._parse_script_name(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(script_name), buf


class RenameScriptCommand(Command):

    command = b'RENAMESCRIPT'

    def __init__(self, old_script_name: str, new_script_name: str) -> None:
        super().__init__()
        self.old_script_name: Final = old_script_name
        self.new_script_name: Final = new_script_name

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[RenameScriptCommand, memoryview]:
        old_script_name, buf = cls._parse_script_name(buf, params)
        new_script_name, buf = cls._parse_script_name(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(old_script_name, new_script_name), buf


class CheckScriptCommand(Command):

    command = b'CHECKSCRIPT'

    def __init__(self, script_data: bytes) -> None:
        super().__init__()
        self.script_data: Final = script_data

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[CheckScriptCommand, memoryview]:
        script_data, buf = String.parse(buf, params)
        _, buf = EndLine.parse(buf, params)
        return cls(script_data.value), buf




from __future__ import annotations

import os.path
from argparse import ArgumentParser, Namespace
from collections.abc import Mapping, AsyncIterator
from concurrent.futures import ThreadPoolExecutor
from contextlib import asynccontextmanager, AsyncExitStack
from datetime import datetime
from typing import Any, Optional, Final

from pysasl.creds import AuthenticationCredentials

from pymap.concurrent import Subsystem
from pymap.config import BackendCapability, IMAPConfig
from pymap.exceptions import AuthorizationFailure, NotSupportedError
from pymap.filter import PluginFilterSet, SingleFilterSet
from pymap.health import HealthStatus
from pymap.interfaces.backend import BackendInterface
from pymap.interfaces.login import LoginInterface, IdentityInterface
from pymap.token import AllTokens
from pymap.user import UserMetadata

from .layout import MaildirLayout
from .mailbox import Message, Maildir, MailboxSet
from ..session import BaseSession

__all__ = ['MaildirBackend', 'Config']


class MaildirBackend(BackendInterface):
    """Defines an on-disk backend that uses :class:`~mailbox.Maildir` for
    mailbox storage and `MailboxFormat/Maildir
    <https://wiki2.dovecot.org/MailboxFormat/Maildir>`_ for metadata storage.

    """

    def __init__(self, login: Login, config: Config) -> None:
        super().__init__()
        self._login = login
        self._config = config
        self._status = HealthStatus()

    @property
    def login(self) -> Login:
        return self._login

    @property
    def users(self) -> None:
        return None

    @property
    def config(self) -> Config:
        return self._config

    @property
    def status(self) -> HealthStatus:
        return self._status

    @classmethod
    def add_subparser(cls, name: str, subparsers: Any) -> ArgumentParser:
        parser = subparsers.add_parser(name, help='on-disk backend')
        parser.add_argument('users_file', help='path the the users file')
        parser.add_argument('--base-dir', metavar='DIR',
                            help='base directory for mailbox relative paths')
        parser.add_argument('--concurrency', metavar='NUM', type=int,
                            help='maximum number of IO workers')
        parser.add_argument('--layout', metavar='TYPE', default='++',
                            help='maildir directory layout')
        return parser

    @classmethod
    async def init(cls, args: Namespace, **overrides: Any) \
            -> tuple[MaildirBackend, Config]:
        config = Config.from_args(args)
        login = Login(config)
        return cls(login, config), config

    async def start(self, stack: AsyncExitStack) -> None:
        pass


class Config(IMAPConfig):
    """The config implementation for the maildir backend.

    Args:
        args: The command-line arguments.
        users_file: The path to the users file.
        base_dir: The base directory for all relative mailbox paths.
        layout: The Maildir directory layout.

    """

    def __init__(self, args: Namespace, *, users_file: str,
                 base_dir: Optional[str], layout: str, **extra: Any) -> None:
        super().__init__(args, **extra)
        self._users_file = users_file
        self._base_dir = self._get_base_dir(base_dir, users_file)
        self._layout = layout

    @classmethod
    def _get_base_dir(cls, base_dir: Optional[str],
                      users_file: Optional[str]) -> str:
        if base_dir:
            return base_dir
        elif users_file:
            return os.path.dirname(users_file)
        else:
            raise ValueError('--base-dir', base_dir)

    @property
    def backend_capability(self) -> BackendCapability:
        return BackendCapability(idle=True, object_id=True, multi_append=True)

    @property
    def users_file(self) -> str:
        """Used by the default :meth:`~Session.find_user` implementation
        to retrieve the users file path from the command-line arguments. The
        users file is given as the first positional argument on the
        command-line.

        This file contains a valid login on each line, which are split into
        three parts by colon (``:``) characters: the user name, the mailbox
        path, and the password.

        The password may contain colon characters. The mailbox path may be
        empty, relative, or absolute. If it is empty, the user ID is used as a
        relative path.

        """
        return self._users_file

    @property
    def base_dir(self) -> str:
        """The base directory for all relative mailbox paths. The default is
        the directory containing the users file.

        """
        return self._base_dir

    @property
    def layout(self) -> str:
        """The Maildir directory layout name.

        See Also:
            :class:`~pymap.backend.maildir.layout.MaildirLayout`

        """
        return self._layout

    @classmethod
    def parse_args(cls, args: Namespace) -> Mapping[str, Any]:
        executor = ThreadPoolExecutor(args.concurrency)
        subsystem = Subsystem.for_executor(executor)
        return {**super().parse_args(args),
                'users_file': args.users_file,
                'base_dir': args.base_dir,
                'layout': args.layout,
                'subsystem': subsystem}


class FilterSet(PluginFilterSet[bytes], SingleFilterSet[bytes]):

    def __init__(self, user_dir: str) -> None:
        super().__init__('sieve', bytes)
        self._user_dir = user_dir

    async def replace_active(self, value: Optional[bytes]) -> None:
        path = os.path.join(self._user_dir, 'dovecot.sieve')
        if value is None:
            try:
                os.unlink(path)
            except FileNotFoundError:
                pass
        else:
            with open(path, 'wb') as sieve_file:
                sieve_file.write(value)

    async def get_active(self) -> Optional[bytes]:
        path = os.path.join(self._user_dir, 'dovecot.sieve')
        try:
            with open(path, 'rb') as sieve_file:
                return sieve_file.read()
        except FileNotFoundError:
            return None


class Session(BaseSession[Message]):
    """The session implementation for the maildir backend."""

    resource = __name__

    def __init__(self, owner: str, config: Config, mailbox_set: MailboxSet,
                 filter_set: FilterSet) -> None:
        super().__init__(owner)
        self._config = config
        self._mailbox_set = mailbox_set
        self._filter_set = filter_set

    @property
    def config(self) -> Config:
        return self._config

    @property
    def mailbox_set(self) -> MailboxSet:
        return self._mailbox_set

    @property
    def filter_set(self) -> FilterSet:
        return self._filter_set


class Login(LoginInterface):
    """The login implementation for the maildir backend."""

    def __init__(self, config: Config) -> None:
        super().__init__()
        self.config = config
        self._tokens = AllTokens()

    @property
    def tokens(self) -> AllTokens:
        return self._tokens

    async def authenticate(self, credentials: AuthenticationCredentials) \
            -> Identity:
        authcid = credentials.authcid
        identity = credentials.identity
        password: Optional[str] = None
        mailbox_path: Optional[str] = None
        with open(self.config.users_file, 'r') as users_file:
            for line in users_file:
                this_user, this_user_dir, this_password = line.split(':', 2)
                if authcid == this_user:
                    password = this_password.rstrip('\r\n')
                if identity == this_user:
                    mailbox_path = this_user_dir or this_user
        data = UserMetadata(self.config, password=password)
        await data.check_password(credentials)
        if mailbox_path is None or authcid != identity:
            raise AuthorizationFailure()
        return Identity(self.config, identity, data, mailbox_path)


class Identity(IdentityInterface):
    """The identity implementation for the maildir backend."""

    def __init__(self, config: Config, name: str, metadata: UserMetadata,
                 mailbox_path: str) -> None:
        super().__init__()
        self.config: Final = config
        self.metadata: Final = metadata
        self.mailbox_path: Final = mailbox_path
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    async def new_token(self, *, expiration: datetime = None) -> None:
        return None

    @asynccontextmanager
    async def new_session(self) -> AsyncIterator[Session]:
        config = self.config
        maildir, layout = self._load_maildir()
        mailbox_set = MailboxSet(maildir, layout)
        filter_set = FilterSet(layout.path)
        yield Session(self.name, config, mailbox_set, filter_set)

    def _load_maildir(self) -> tuple[Maildir, MaildirLayout]:
        full_path = os.path.join(self.config.base_dir, self.mailbox_path)
        layout = MaildirLayout.get(full_path, self.config.layout, Maildir)
        create = not os.path.exists(full_path)
        return Maildir(full_path, create=create), layout

    async def get(self) -> UserMetadata:
        return self.metadata

    async def set(self, metadata: UserMetadata) -> None:
        raise NotSupportedError()

    async def delete(self) -> None:
        raise NotSupportedError()



"""Defines useful types and utilities for working with bytestrings."""

from __future__ import annotations

import zlib
from abc import abstractmethod, ABCMeta
from collections.abc import Iterable, Sequence
from io import BytesIO
from itertools import chain
from typing import cast, final, Any, Final, TypeVar, SupportsBytes, Union, \
    Protocol

__all__ = ['MaybeBytes', 'MaybeBytesT', 'WriteStream', 'Writeable',
           'BytesFormat']

#: A bytes object, memoryview,  or an object with a ``__bytes__`` method.
MaybeBytes = Union[bytes, bytearray, memoryview, SupportsBytes]

#: A type variable bound to :class:`MaybeBytes`.
MaybeBytesT = TypeVar('MaybeBytesT', bound=MaybeBytes)

_FormatArg = Union[MaybeBytes, int]


class WriteStream(Protocol):
    """Typing protocol indicating the object implements the :meth:`.write`
    method.

    See Also:
        :class:`~asyncio.StreamWriter`, :class:`~typing.BinaryIO`

    """

    @abstractmethod
    def write(self, data: bytes) -> Any:
        """Defines an abstract method where ``data`` is written to a stream or
        buffer.

        Args:
            data: The data to write.

        """
        ...


class HashStream(WriteStream):
    """A stream that a :class:`Writeable` can use to generate a
    non-cryptographic hash using :func:`zlib.adler32`.

    """

    __slots__ = ['_digest']

    def __init__(self) -> None:
        super().__init__()
        self._digest = zlib.adler32(b'')

    def write(self, data: bytes) -> None:
        self._digest = zlib.adler32(data, self._digest)

    def digest(self, data: Writeable = None) -> bytes:
        """Return the digest of the data written to the hash stream.

        Args:
            data: The data to write before computing the digest.

        """
        if data is not None:
            data.write(self)
        return self._digest.to_bytes(4, 'big')


class Writeable(metaclass=ABCMeta):
    """Base class for types that can be written to a stream."""

    __slots__: Sequence[str] = []

    @final
    def tobytes(self) -> bytes:
        """Convert the writeable object back into a bytestring using the
        :meth:`.write` method.

        """
        writer = BytesIO()
        self.write(writer)
        return writer.getvalue()

    @classmethod
    def empty(cls) -> Writeable:
        """Return a :class:`Writeable` for an empty string."""
        return _EmptyWriteable()

    @classmethod
    def wrap(cls, data: MaybeBytes) -> Writeable:
        """Wrap the bytes in a :class:`Writeable`.

        Args:
            data: The object to wrap.

        """
        return _WrappedWriteable(data)

    @classmethod
    def concat(cls, data: Iterable[MaybeBytes]) -> Writeable:
        """Wrap the iterable in a :class:`Writeable` that will write each item.

        Args:
            data: The iterable to wrap.

        """
        return _ConcatWriteable(data)

    def write(self, writer: WriteStream) -> None:
        """Write the object to the stream, with one or more calls to
        :meth:`~WriteStream.write`.

        Args:
            writer: The output stream.

        """
        writer.write(bytes(self))

    def __bool__(self) -> bool:
        return True

    def __len__(self) -> int:
        return len(bytes(self))

    @abstractmethod
    def __bytes__(self) -> bytes:
        ...


class _EmptyWriteable(Writeable):

    __slots__: Sequence[str] = []

    def write(self, writer: WriteStream) -> None:
        pass

    def __bytes__(self) -> bytes:
        return b''

    def __repr__(self) -> str:
        return '<Writeable empty>'


class _WrappedWriteable(Writeable):

    __slots__ = ['data']

    def __init__(self, data: MaybeBytes) -> None:
        if isinstance(data, bytes):
            self.data = data
        else:
            self.data = bytes(data)

    def __bytes__(self) -> bytes:
        return self.data

    def __repr__(self) -> str:
        return f'<Writeable {repr(self.data)}>'


class _ConcatWriteable(Writeable):

    __slots__ = ['data']

    def __init__(self, data: Iterable[MaybeBytes]) -> None:
        self.data = list(data)

    def write(self, writer: WriteStream) -> None:
        for item in self.data:
            if isinstance(item, Writeable):
                item.write(writer)
            else:
                writer.write(bytes(item))

    def __bytes__(self) -> bytes:
        return BytesFormat(b'').join(self.data)

    def __repr__(self) -> str:
        return f'<Writeable {repr(self.data)}>'


class BytesFormat:
    """Helper utility for performing formatting operations that produce
    bytestrings. While similar to the builtin formatting and join
    operations, this class intends to provide cleaner typing.

    Args:
        how: The formatting string or join delimiter to use.

    """

    __slots__ = ['how']

    def __init__(self, how: bytes) -> None:
        super().__init__()
        self.how: Final = how

    def __mod__(self, other: Union[_FormatArg, Iterable[_FormatArg]]) -> bytes:
        """String interpolation, shortcut for :meth:`.format`.

        Args:
            other: The data interpolated into the format string.

        """
        if isinstance(other, bytes):
            return self.format([other])
        elif hasattr(other, '__bytes__'):
            supports_bytes = cast(SupportsBytes, other)
            return self.format([bytes(supports_bytes)])
        elif hasattr(other, '__iter__'):
            items = cast(Iterable[_FormatArg], other)
            return self.format(items)
        return NotImplemented

    @classmethod
    def _fix_format_arg(cls, data: _FormatArg) -> Any:
        if isinstance(data, int):
            return data
        else:
            return bytes(data)

    def format(self, data: Iterable[_FormatArg]) -> bytes:
        """String interpolation into the format string.

        Args:
            data: The data interpolated into the format string.

        Examples:
            ::

                BytesFormat(b'Hello, %b!') % b'World'
                BytesFormat(b'%b, %b!') % (b'Hello', b'World')

        """
        fix_arg = self._fix_format_arg
        return self.how % tuple(fix_arg(item) for item in data)

    @classmethod
    def _fix_join_arg(cls, data: _FormatArg) -> Any:
        if isinstance(data, int):
            return b'%d' % data
        else:
            return bytes(data)

    def join(self, *data: Iterable[_FormatArg]) -> bytes:
        """Iterable join on a delimiter.

        Args:
            data: Iterable of items to join.

        Examples:
            ::

                BytesFormat(b' ').join([b'one', b'two', b'three'])

        """
        fix_arg = self._fix_join_arg
        return self.how.join(fix_arg(item) for item in chain(*data))



"""Primitive parseable objects in the IMAP protocol."""

from __future__ import annotations

import re
from abc import abstractmethod, ABCMeta
from collections.abc import Iterable, Iterator, Sequence
from functools import total_ordering
from re import Match
from typing import cast, Union, Optional, SupportsBytes

from . import Parseable, ExpectedParseable, Params
from .exceptions import NotParseable
from .state import ExpectContinuation
from ..bytes import rev, MaybeBytes, MaybeBytesT, BytesFormat, WriteStream, \
    Writeable

__all__ = ['Nil', 'Number', 'Atom', 'List', 'String',
           'QuotedString', 'LiteralString']


class Nil(Parseable[None]):
    """Represents a ``NIL`` object from an IMAP stream."""

    _nil_pattern = rev.compile(b'^NIL$', re.I)

    __slots__: list[str] = []

    def __init__(self) -> None:
        super().__init__()

    @property
    def value(self) -> None:
        """Always returns ``None``."""
        return None

    def __bytes__(self) -> bytes:
        return b'NIL'

    def __hash__(self) -> int:
        return hash(Nil)

    def __eq__(self, other) -> bool:
        if isinstance(other, Nil):
            return True
        return super().__eq__(other)

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[Nil, memoryview]:
        start = cls._whitespace_length(buf)
        match = cls._atom_pattern.match(buf, start)
        if not match:
            raise NotParseable(buf)
        atom = match.group(0)
        if not cls._nil_pattern.match(atom):
            raise NotParseable(buf)
        return cls(), buf[match.end(0):]


@total_ordering
class Number(Parseable[int]):
    """Represents a number object from an IMAP stream.

    Args:
        num: The integer value.

    """

    _num_pattern = rev.compile(br'^\d+$')

    __slots__ = ['num', '_raw']

    def __init__(self, num: int) -> None:
        super().__init__()
        self.num = num
        self._raw = b'%d' % num

    @property
    def value(self) -> int:
        """The integer value."""
        return self.num

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[Number, memoryview]:
        start = cls._whitespace_length(buf)
        match = cls._atom_pattern.match(buf, start)
        if not match:
            raise NotParseable(buf)
        atom = match.group(0)
        if not cls._num_pattern.match(atom):
            raise NotParseable(buf)
        return cls(int(match.group(0))), buf[match.end(0):]

    def __bytes__(self) -> bytes:
        return self._raw

    def __hash__(self) -> int:
        return hash((Number, self.value))

    def __eq__(self, other) -> bool:
        if isinstance(other, Number):
            return self.value == other.value
        elif isinstance(other, int):
            return self.value == other
        return super().__eq__(other)

    def __lt__(self, other) -> bool:
        if isinstance(other, Number):
            return self.value < other.value
        elif isinstance(other, int):
            return self.value < other
        return NotImplemented


class Atom(Parseable[bytes]):
    """Represents an atom object from an IMAP stream.

    Args:
        value: The atom bytestring.

    """

    __slots__ = ['_value']

    def __init__(self, value: bytes) -> None:
        super().__init__()
        self._value = value

    @property
    def value(self) -> bytes:
        """The atom bytestring."""
        return self._value

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[Atom, memoryview]:
        start = cls._whitespace_length(buf)
        match = cls._atom_pattern.match(buf, start)
        if not match:
            raise NotParseable(buf[start:])
        atom = match.group(0)
        return cls(atom), buf[match.end(0):]

    def __bytes__(self) -> bytes:
        return self.value

    def __hash__(self) -> int:
        return hash((Atom, self.value))

    def __eq__(self, other) -> bool:
        if isinstance(other, Atom):
            return self.value == other.value
        return super().__eq__(other)


class String(Parseable[bytes], metaclass=ABCMeta):
    """Represents a string object from an IMAP string. This object may not be
    instantiated directly, use one of its derivatives instead.

    """

    _MAX_LEN = 4096

    __slots__: list[str] = []

    @property
    @abstractmethod
    def binary(self) -> bool:
        """True if the string should be transmitted as binary."""
        ...

    @property
    @abstractmethod
    def length(self) -> int:
        """The length of the string value."""
        ...

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[String, memoryview]:
        try:
            return QuotedString.parse(buf, params)
        except NotParseable:
            pass
        return LiteralString.parse(buf, params)

    @classmethod
    def build(cls, value: object, binary: bool = False,
              fallback: object = None) -> Union[Nil, String]:
        """Produce either a :class:`QuotedString` or :class:`LiteralString`
        based on the contents of ``data``. This is useful to improve
        readability of response data.

        Args:
            value: The string to serialize.
            binary: True if the string should be transmitted as binary.
            fallback: The default value to use if ``value`` is None.

        """
        if value is None:
            if fallback is None:
                return Nil()
            else:
                return cls.build(fallback, binary)
        elif not value:
            return QuotedString(b'')
        elif isinstance(value, bytes):
            ascii_ = value
        elif isinstance(value, memoryview):
            ascii_ = bytes(value)
        elif hasattr(value, '__bytes__'):
            ascii_ = bytes(cast(SupportsBytes, value))
        elif isinstance(value, str) or hasattr(value, '__str__'):
            value = str(value)
            try:
                ascii_ = bytes(value, 'ascii')
            except UnicodeEncodeError:
                ascii_ = bytes(value, 'utf-8', 'replace')
                return LiteralString(ascii_, binary)
        else:
            raise TypeError(value)
        if not binary and len(ascii_) < 64 \
                and b'\n' not in ascii_ \
                and b'\x00' not in ascii_:
            return QuotedString(ascii_)
        else:
            return LiteralString(ascii_, binary)

    def __bytes__(self) -> bytes:
        raise NotImplementedError

    def __hash__(self) -> int:
        return hash((String, self.value))

    def __eq__(self, other) -> bool:
        if isinstance(other, String):
            return self.value == other.value
        return super().__eq__(other)


class QuotedString(String):
    """Represents a string object from an IMAP stream that was encased in
    double-quotes.

    Args:
        string: The string value.

    """

    _quoted_pattern = rev.compile(br'(?:\r|\n|\\.|\")')
    _quoted_specials_pattern = rev.compile(br'[\"\\]')

    __slots__ = ['_string', '_raw']

    def __init__(self, string: bytes, raw: bytes = None) -> None:
        super().__init__()
        self._string = string
        self._raw = raw

    @property
    def value(self) -> bytes:
        """The string value."""
        return self._string

    @property
    def binary(self) -> bool:
        return False

    @property
    def length(self) -> int:
        return len(self._string)

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[QuotedString, memoryview]:
        start = cls._whitespace_length(buf)
        if buf[start:start + 1] != b'"':
            raise NotParseable(buf)
        marker = start + 1
        unquoted = bytearray()
        for match in cls._quoted_pattern.finditer(buf, marker):
            unquoted += buf[marker:match.start(0)]
            match_group = match.group(0)
            if match_group in (b'\r', b'\n'):
                raise NotParseable(buf)
            elif match_group.startswith(b'\\'):
                escape_char = match_group[-1:]
                if escape_char in (b'\\', b'"'):
                    unquoted += escape_char
                else:
                    raise NotParseable(buf)
                marker = match.end(0)
            else:
                end = match.end(0)
                quoted = buf[start:end + 1]
                return cls(bytes(unquoted), bytes(quoted)), buf[end:]
        raise NotParseable(buf)

    @classmethod
    def _escape_quoted_specials(cls, match: Match) -> bytes:
        return b'\\' + match.group(0)

    def __bytes__(self) -> bytes:
        if self._raw is not None:
            return bytes(self._raw)
        pat = self._quoted_specials_pattern
        quoted_string = pat.sub(self._escape_quoted_specials, self.value)
        self._raw = BytesFormat(b'"%b"') % (quoted_string, )
        return self._raw


class LiteralString(String):
    """Represents a string object from an IMAP stream that used the literal
    syntax.

    Args:
        string: The literal string value.
        binary: True if the string is considered binary data.

    """

    _literal_pattern = rev.compile(br'(~?){(\d+)(\+?)}\r?\n')

    __slots__ = ['_string', '_length', '_binary', '_raw']

    def __init__(self, string: Union[bytes, Writeable],
                 binary: bool = False) -> None:
        super().__init__()
        self._string = string
        self._length = len(string)
        self._binary = binary
        self._raw: Optional[bytes] = None

    @property
    def value(self) -> bytes:
        return bytes(self._string)

    @property
    def binary(self) -> bool:
        return self._binary

    @property
    def length(self) -> int:
        return self._length

    @property
    def _prefix(self) -> bytes:
        binary_prefix = b'~' if self.binary else b''
        return b'%b{%d}\r\n' % (binary_prefix, self.length)

    @classmethod
    def _check_too_big(cls, params: Params, length: int) -> bool:
        if params.command_name == b'APPEND':
            max_len = params.max_append_len
        else:
            max_len = cls._MAX_LEN
        return max_len is not None and length > max_len

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[LiteralString, memoryview]:
        start = cls._whitespace_length(buf)
        match = cls._literal_pattern.match(buf, start)
        if not match:
            raise NotParseable(buf)
        binary = match.group(1) == b'~'
        literal_length = int(match.group(2))
        if cls._check_too_big(params, literal_length):
            raise NotParseable(buf, b'TOOBIG')
        elif match.group(3) == b'+':
            buf = buf[match.end(0):]
            literal = bytes(buf[0:literal_length])
        elif len(buf) > match.end(0):
            raise NotParseable(buf[match.end(0):])
        elif params.allow_continuations:
            expected = ExpectContinuation(b'Literal string', literal_length)
            buf = expected.expect(params.state)
            literal = bytes(buf[0:literal_length])
        else:
            raise NotParseable(buf)
        if len(literal) != literal_length:
            raise NotParseable(buf)
        return cls(literal, binary), buf[literal_length:]

    def write(self, writer: WriteStream) -> None:
        writer.write(self._prefix)
        if isinstance(self._string, Writeable):
            self._string.write(writer)
        else:
            writer.write(self._string)

    def __len__(self) -> int:
        return len(self._prefix) + self.length

    def __bytes__(self) -> bytes:
        if self._raw is None:
            self._raw = self.tobytes()
        return self._raw


class List(Parseable[Sequence[MaybeBytes]]):
    """Represents a list of :class:`Parseable` objects from an IMAP stream.

    Args:
        items: The list of parsed objects.
        sort: If True, the list of items is sorted.

    """

    _end_pattern = rev.compile(br' *\)')

    __slots__ = ['items']

    def __init__(self, items: Iterable[MaybeBytes],
                 sort: bool = False) -> None:
        super().__init__()
        if sort:
            items_list = sorted(items)  # type: ignore
        else:
            items_list = list(items)
        self.items: Sequence[MaybeBytes] = items_list

    @property
    def value(self) -> Sequence[MaybeBytes]:
        """The list of parsed objects."""
        return self.items

    def get_as(self, cls: type[MaybeBytesT]) -> Sequence[MaybeBytesT]:
        """Return the list of parsed objects."""
        _ = cls  # noqa
        return cast(Sequence[MaybeBytesT], self.items)

    def __iter__(self) -> Iterator[MaybeBytes]:
        return iter(self.value)

    def __len__(self) -> int:
        return len(self.value)

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[List, memoryview]:
        start = cls._whitespace_length(buf)
        if buf[start:start + 1] != b'(':
            raise NotParseable(buf)
        items: list[Parseable] = []
        buf = buf[start + 1:]
        while True:
            match = cls._end_pattern.match(buf)
            if match:
                return cls(items), buf[match.end(0):]
            elif items and not cls._whitespace_length(buf):
                raise NotParseable(buf)
            params_copy = params.copy(expected=params.list_expected)
            item, buf = ExpectedParseable.parse(buf, params_copy)
            items.append(item)

    def write(self, writer: WriteStream) -> None:
        writer.write(b'(')
        is_first = True
        for i, item in enumerate(self.items):
            if is_first:
                is_first = False
            else:
                writer.write(b' ')
            if isinstance(item, Writeable):
                item.write(writer)
            else:
                writer.write(bytes(item))
        writer.write(b')')

    def __bytes__(self) -> bytes:
        raw_items = BytesFormat(b' ').join(self.items)
        return b'(%b)' % raw_items

    def __hash__(self) -> int:
        return hash((List, self.value))

    def __eq__(self, other) -> bool:
        if isinstance(other, List):
            return self.__eq__(other.value)
        elif isinstance(other, Sequence):
            if len(self.value) != len(other):
                return False
            for i, val in enumerate(self.value):
                if val != other[i]:
                    return False
            return True
        return super().__eq__(other)




from __future__ import annotations

from collections.abc import Callable, Iterable, Iterator, Mapping
from typing import TypeVar, Generic, Optional, Final

from pkg_resources import iter_entry_points, DistributionNotFound

__all__ = ['PluginT', 'Plugin']

#: The plugin type variable.
PluginT = TypeVar('PluginT')


class Plugin(Generic[PluginT], Iterable[tuple[str, PluginT]]):
    """Plugin system, typically loaded from :mod:`pkg_resources` `entry points
    <https://packaging.python.org/guides/creating-and-discovering-plugins/#using-package-metadata>`_.

    >>> example: Plugin[type[Example]] = Plugin('plugins.example')
    >>> example.add('two', ExampleTwo)
    >>> example.registered
    {'one': <class 'examples.ExampleOne'>,
     'two': <class 'examples.ExampleTwo'>}

    Note:
        Plugins registered from *group* entry points are lazy-loaded. This is
        to avoid cyclic imports.

    Args:
        group: The entry point group to load.

    """

    def __init__(self, group: str, *, default: str = None) -> None:
        super().__init__()
        self.group: Final = group
        self._default = default
        self._loaded: Optional[dict[str, PluginT]] = None
        self._added: dict[str, PluginT] = {}

    def __iter__(self) -> Iterator[tuple[str, PluginT]]:
        return iter(self.registered.items())

    @property
    def registered(self) -> Mapping[str, PluginT]:
        """A mapping of the registered plugins, keyed by name."""
        loaded = self._load()
        return {**loaded, **self._added}

    @property
    def default(self) -> PluginT:
        """The default plugin implementation.

        This property may also be assigned a new string value to change the
        name of the default plugin.

        >>> example: Plugin[type[Example]] = Plugin('plugins.example',
        ...                                         default='one')
        >>> example.default
        <class 'examples.ExampleOne'>
        >>> example.default = 'two'
        >>> example.default
        <class 'examples.ExampleTwo'>

        Raises:
            KeyError: The default plugin name was not registered.

        """
        if self._default is None:
            raise KeyError(f'{self.group!r} has no default plugin')
        else:
            return self.registered[self._default]

    @default.setter
    def default(self, default: Optional[str]) -> None:
        self._default = default

    def _load(self) -> Mapping[str, PluginT]:
        loaded = self._loaded
        if loaded is None:
            loaded = {}
            for entry_point in iter_entry_points(self.group):
                try:
                    plugin: PluginT = entry_point.load()
                except DistributionNotFound:
                    pass  # optional dependencies not installed
                else:
                    loaded[entry_point.name] = plugin
            self._loaded = loaded
        return loaded

    def add(self, name: str, plugin: PluginT) -> None:
        """Add a new plugin by name.

        Args:
            name: The identifying name of the plugin.
            plugin: The plugin object.

        """
        self._added[name] = plugin

    def register(self, name: str) -> Callable[[PluginT], PluginT]:
        """Decorates a plugin implementation.

        Args:
            name: The identifying name of the plugin.

        """
        def deco(plugin: PluginT) -> PluginT:
            self.add(name, plugin)
            return plugin
        return deco

    def __repr__(self) -> str:
        return f'Plugin({self.group!r})'




from __future__ import annotations

from abc import abstractmethod
from collections.abc import Set
from datetime import datetime
from typing import Optional, Protocol

from pysasl.creds import AuthenticationCredentials

__all__ = ['TokensInterface']


class TokensInterface(Protocol):
    """Defines the create and parse operations for a token type."""

    @abstractmethod
    def get_login_token(self, identifier: str, key: bytes, *,
                        authzid: str = None, location: str = None,
                        expiration: datetime = None) -> Optional[str]:
        """Returns a new token string that encapsulates the provided login
        data, or ``None`` if tokens are not supported.

        Args:
            identifier: Any string that can facilitate the lookup of *key*.
            key: The private key used to create and verify the token.
            authzid: Limits the token to authorizing as this identity.
            location: Arbitrary metadata string, application-specific.
            expiration: An expiration when this token shall no longer be valid.

        """
        ...

    @abstractmethod
    def get_admin_token(self, admin_key: Optional[bytes], *,
                        authzid: str = None, location: str = None,
                        expiration: datetime = None) -> Optional[str]:
        """Returns a new token string that encapsulates the provided login
        data, or ``None`` if tokens are not supported. This token uses a
        special private admin key that grants unrestricted access.

        Args:
            admin_key: The private admin key.
            authzid: Limits the token to authorizing as this identity.
            location: Arbitrary metadata string, application-specific.
            expiration: An expiration when this token shall no longer be valid.

        """
        ...

    @abstractmethod
    def parse(self, authzid: str, token: str, *,
              admin_keys: Set[bytes] = ...) -> AuthenticationCredentials:
        """Parses a token string to produce authentication credentials.

        Args:
            authzid: Check token for authorization to this identity.
            token: The token string.
            admin_keys: The set of possible private admin keys.

        Raises:
            ValueError: The *token* string was not valid.

        """
        ...




from __future__ import annotations

import asyncio
import logging
import uuid
from abc import abstractmethod
from asyncio import Task, CancelledError
from contextlib import AsyncExitStack
from typing import ClassVar, Protocol

from aioredis import Redis, ConnectionError
from pymap.context import connection_exit
from pymap.health import HealthStatus

__all__ = ['BackgroundAction', 'BackgroundTask', 'NoopAction']

_log = logging.getLogger(__name__)


class BackgroundAction(Protocol):
    """An action to perform with the connection."""

    @abstractmethod
    async def __call__(self, conn: Redis, duration: float) -> None:
        ...


class NoopAction(BackgroundAction):
    """A background action that calls :meth:`~aioredis.client.Redis.blpop` on
    a randomly-generated key.

    """

    async def __call__(self, conn: Redis, duration: float) -> None:
        key = b'invalid-' + uuid.uuid4().bytes + uuid.uuid4().bytes
        duration2: int = duration  # type: ignore
        await conn.blpop(key, timeout=duration2)


class BackgroundTask:
    """Maintains a :class:`CleanupThread` for the duration of the process
    lifetime, restarting on failure.

    Args:
        address: The redis server address for mail data.
        status: The system health status.
        root: The root redis key.

    """

    #: The delay between redis reconnect attempts, on connection failure.
    connection_delay: ClassVar[float] = 5.0

    #: The duration to hold an active connection before reconnecting, to detect
    #: silent failures.
    reconnect_duration: ClassVar[float] = 30.0

    def __init__(self, redis: Redis, status: HealthStatus,
                 action: BackgroundAction) -> None:
        super().__init__()
        self._redis = redis
        self._status = status
        self._action = action

    async def _run_forever(self) -> None:
        while True:
            try:
                async with AsyncExitStack() as stack:
                    connection_exit.set(stack)
                    conn = await stack.enter_async_context(
                        self._redis.client())
                    self._status.set_healthy()
                    await self._action(conn, self.reconnect_duration)
            except (ConnectionError, OSError):
                self._status.set_unhealthy()
            except CancelledError:
                break
            await asyncio.sleep(self.connection_delay)

    def start(self) -> Task[None]:
        """Return a task running the cleanup loop indefinitely."""
        return asyncio.create_task(self._run_forever())




from __future__ import annotations

from grpclib.const import Status
from grpclib.exceptions import GRPCError
from google.rpc.error_details_pb2 import ErrorInfo

__all__ = ['get_unimplemented_error', 'get_incompatible_version_error']


def get_unimplemented_error(*, domain: str = None, **metadata: str) \
        -> GRPCError:
    """Build a :exc:`~grpclib.exceptions.GRPCError` exception for an
    operation that is not implemented by the server.

    Args:
        domain: The domain string to include in the error.
        metadata: Additional metadata to include in the error.

    """
    return GRPCError(Status.UNIMPLEMENTED, 'Operation not available', [
        ErrorInfo(reason='UNIMPLEMENTED', domain=domain, metadata=metadata)])


def get_incompatible_version_error(client_version: str, server_version: str, *,
                                   domain: str = None, **metadata: str) \
        -> GRPCError:
    """Build a :exc:`~grpclib.exceptions.GRPCError` exception for an
    incompatible version error.

    Args:
        client_version: The client version string.
        server_version: The server version string.
        domain: The domain string to include in the error.
        metadata: Additional metadata to include in the error.

    """
    metadata = {'client_version': client_version,
                'server_version': server_version,
                **metadata}
    return GRPCError(Status.FAILED_PRECONDITION, 'Incompatible version', [
        ErrorInfo(reason='INCOMPATIBLE', domain=domain, metadata=metadata)])




import unittest
from datetime import datetime

from pymap.flags import FlagOp, PermanentFlags, SessionFlags
from pymap.message import BaseMessage
from pymap.parsing.command.select import SearchCommand, UidSearchCommand
from pymap.parsing.response import ResponseOk
from pymap.parsing.specials import SequenceSet, ObjectId
from pymap.parsing.specials.flag import Seen, Flagged, Flag
from pymap.selected import SelectedMailbox

_Keyword = Flag(b'$Keyword')


class _Message(BaseMessage):

    async def load_content(self, requirement):
        raise RuntimeError()


class TestSelectedMailbox(unittest.TestCase):

    def setUp(self) -> None:
        self.response = ResponseOk(b'.', b'testing')

    @classmethod
    def new_selected(cls, guid: bytes = b'test') -> SelectedMailbox:
        return SelectedMailbox(ObjectId(guid), False,
                               PermanentFlags([Seen, Flagged]),
                               SessionFlags([_Keyword]))

    @classmethod
    def set_messages(cls, selected: SelectedMailbox,
                     expunged, messages) -> None:
        updates = [_Message(uid, datetime.now(), flags)
                   for uid, flags in messages]
        selected.add_updates(updates, expunged)

    @property
    def command(self) -> SearchCommand:
        return SearchCommand(b'.', [], None)

    @property
    def uid_command(self) -> SearchCommand:
        return UidSearchCommand(b'.', [], None)

    def test_add_untagged_recent_equal(self) -> None:
        selected = self.new_selected()
        selected.session_flags.add_recent(1)
        selected.session_flags.add_recent(2)
        self.set_messages(selected, [],
                          [(1, []), (2, [])])
        forked, _ = selected.fork(self.command)
        self.set_messages(forked, [],
                          [(1, []), (2, [])])
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_recent_increase(self) -> None:
        selected = self.new_selected()
        selected.session_flags.add_recent(1)
        selected.session_flags.add_recent(2)
        self.set_messages(selected, [],
                          [(1, []), (2, []), (3, [])])
        forked, _ = selected.fork(self.command)
        forked.session_flags.add_recent(3)
        self.set_messages(forked, [], [])
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 3 RECENT\r\n'
                         b'* 3 FETCH (FLAGS (\\Recent))\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_session_flag_add(self) -> None:
        selected = self.new_selected()
        self.set_messages(selected, [],
                          [(1, []), (2, [Seen]), (3, [])])
        forked, _ = selected.fork(self.command)
        forked.session_flags.update(2, [_Keyword], FlagOp.ADD)
        self.set_messages(forked, [], [])
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 2 FETCH (FLAGS (\\Seen $Keyword))\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_recent_expunge(self) -> None:
        selected = self.new_selected()
        selected.session_flags.add_recent(1)
        selected.session_flags.add_recent(2)
        self.set_messages(selected, [],
                          [(1, []), (2, []), (3, [])])
        forked, _ = selected.fork(self.command)
        self.set_messages(forked, [2, 3],
                          [(1, [])])
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 3 EXPUNGE\r\n'
                         b'* 2 EXPUNGE\r\n'
                         b'* 1 RECENT\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_equal(self) -> None:
        selected = self.new_selected()
        self.set_messages(selected, [],
                          [(1, []), (2, [])])
        forked, _ = selected.fork(self.command)
        self.set_messages(forked, [],
                          [(1, []), (2, [])])
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_fetch(self) -> None:
        selected = self.new_selected()
        self.set_messages(selected, [],
                          [(1, []), (2, [])])
        forked, _ = selected.fork(self.command)
        self.set_messages(forked, [],
                          [(2, [Seen]), (3, [Seen, Flagged]), (4, [])])
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 4 EXISTS\r\n'
                         b'* 2 FETCH (FLAGS (\\Seen))\r\n'
                         b'* 3 FETCH (FLAGS (\\Flagged \\Seen))\r\n'
                         b'* 4 FETCH (FLAGS ())\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_fetch_uid(self) -> None:
        selected = self.new_selected()
        self.set_messages(selected, [],
                          [(1, []), (2, [])])
        forked, _ = selected.fork(self.uid_command)
        self.set_messages(forked, [],
                          [(2, [Seen]), (3, [Seen, Flagged]), (4, [])])
        _, untagged = forked.fork(self.uid_command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 4 EXISTS\r\n'
                         b'* 2 FETCH (FLAGS (\\Seen) UID 2)\r\n'
                         b'* 3 FETCH (FLAGS (\\Flagged \\Seen) UID 3)\r\n'
                         b'* 4 FETCH (FLAGS () UID 4)\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_fetch_silenced(self) -> None:
        selected = self.new_selected()
        self.set_messages(selected, [],
                          [(1, []), (2, []), (3, [Seen, Flagged]),
                           (4, [Seen, Flagged]), (5, [Flagged]), (6, [Seen])])
        forked, _ = selected.fork(self.uid_command)
        forked.silence(SequenceSet.build([1, 2]),
                       frozenset([Seen]), FlagOp.ADD)
        forked.silence(SequenceSet.build([3, 4]),
                       frozenset([Seen]), FlagOp.DELETE)
        forked.silence(SequenceSet.build([5, 6]),
                       frozenset([Seen]), FlagOp.REPLACE)
        self.set_messages(forked, [],
                          [(1, [Seen, Flagged]), (2, [Seen]), (3, []),
                           (4, [Flagged]), (5, [Seen, Flagged]), (6, [Seen])])
        _, untagged = forked.fork(self.uid_command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 1 FETCH (FLAGS (\\Flagged \\Seen) UID 1)\r\n'
                         b'* 3 FETCH (FLAGS () UID 3)\r\n'
                         b'* 5 FETCH (FLAGS (\\Flagged \\Seen) UID 5)\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_expunge_hidden(self) -> None:
        selected = self.new_selected()
        self.set_messages(selected, [],
                          [(1, []), (2, []), (3, []), (4, [])])
        forked, _ = selected.fork(self.command)
        forked.hide_expunged = True
        self.set_messages(forked, [2, 3],
                          [(5, [Flagged])])
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 5 EXISTS\r\n'
                         b'* 5 FETCH (FLAGS (\\Flagged))\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_expunge(self) -> None:
        selected = self.new_selected()
        self.set_messages(selected, [],
                          [(1, []), (2, []), (3, []), (4, [])])
        forked, _ = selected.fork(self.command)
        self.set_messages(forked, [2, 3],
                          [(5, [])])
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 3 EXPUNGE\r\n'
                         b'* 2 EXPUNGE\r\n'
                         b'* 3 EXISTS\r\n'
                         b'* 3 FETCH (FLAGS ())\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_all(self) -> None:
        selected = self.new_selected()
        self.set_messages(selected, [],
                          [(1, [Flagged]), (2, []), (3, [])])
        forked, _ = selected.fork(self.uid_command)
        selected.session_flags.add_recent(6)
        self.set_messages(forked, [2, 3],
                          [(1, [Seen, Flagged]), (4, [Seen]), (5, [Seen]),
                           (6, [Flagged]), (7, [Seen])])
        _, untagged = forked.fork(self.uid_command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 3 EXPUNGE\r\n'
                         b'* 2 EXPUNGE\r\n'
                         b'* 5 EXISTS\r\n'
                         b'* 1 RECENT\r\n'
                         b'* 1 FETCH (FLAGS (\\Flagged \\Seen) UID 1)\r\n'
                         b'* 2 FETCH (FLAGS (\\Seen) UID 4)\r\n'
                         b'* 3 FETCH (FLAGS (\\Seen) UID 5)\r\n'
                         b'* 4 FETCH (FLAGS (\\Flagged \\Recent) UID 6)\r\n'
                         b'* 5 FETCH (FLAGS (\\Seen) UID 7)\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_deleted_bye(self) -> None:
        selected = self.new_selected()
        self.set_messages(selected, [],
                          [(1, [])])
        forked, _ = selected.fork(self.command)
        self.set_messages(forked, [1],
                          [(2, [])])
        forked.set_deleted()
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* BYE Selected mailbox no longer exists.\r\n'
                         b'. OK testing\r\n', bytes(self.response))




from __future__ import annotations

from abc import abstractmethod, ABCMeta
from typing import Optional, ClassVar

from .. import Params, Parseable, EndLine

__all__ = ['Command', 'CommandNoArgs', 'CommandAny', 'CommandAuth',
           'CommandNonAuth', 'CommandSelect']


class Command(Parseable[bytes], metaclass=ABCMeta):
    """Base class to represent the commands available to clients.

    Args:
        tag: The tag parsed from the beginning of the command line.

    """

    #: The command key, e.g. ``b'NOOP'``.
    command: ClassVar[bytes] = b''

    #: If given, execution of this command is handled by the delegate command.
    delegate: ClassVar[Optional[type[Command]]] = None

    #: True if the command is part of a compound command.
    compound: ClassVar[bool] = False

    def __init__(self, tag: bytes) -> None:
        super().__init__()

        #: The tag parsed from the beginning of the command line.
        self.tag = tag

    @property
    def value(self) -> bytes:
        """The command name."""
        return self.command

    @classmethod
    @abstractmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[Command, memoryview]:
        ...

    def __bytes__(self) -> bytes:
        return b' '.join((self.tag, self.command))

    def __repr__(self) -> str:
        return '<%s tag=%r>' % (type(self).__name__, self.tag)


class CommandNoArgs(Command, metaclass=ABCMeta):
    """Convenience class used to fail parsing when args are given to a command
    that expects nothing.

    """

    @classmethod
    def parse(cls, buf: memoryview, params: Params) \
            -> tuple[CommandNoArgs, memoryview]:
        _, buf = EndLine.parse(buf, params)
        return cls(params.tag), buf


class CommandAny(Command, metaclass=ABCMeta):
    """Represents a command available at any stage of the IMAP session.

    """
    pass


class CommandAuth(Command, metaclass=ABCMeta):
    """Represents a command available when the IMAP session has been
    authenticated.

    """
    pass


class CommandNonAuth(Command, metaclass=ABCMeta):
    """Represents a command available only when the IMAP session has not yet
    authenticated.

    """
    pass


class CommandSelect(CommandAuth, metaclass=ABCMeta):
    """Represents a command available only when the IMAP session has been
    authenticated and a mailbox has been selected.

    """
    pass




from __future__ import annotations

from abc import abstractmethod, ABCMeta
from collections.abc import Iterator, Mapping, Sequence, AsyncIterator
from contextlib import contextmanager, asynccontextmanager
from typing import ClassVar, Optional, Final, Protocol

from .bytes import BytesFormat, MaybeBytes, Writeable
from .interfaces.message import MessageInterface, LoadedMessageInterface
from .parsing.primitives import Nil, Number, List, LiteralString
from .parsing.specials import DateTime, FetchRequirement, FetchAttribute, \
    FetchValue
from .selected import SelectedMailbox

__all__ = ['LoadedMessageProvider', 'DynamicFetchValue',
           'DynamicLoadedFetchValue', 'MessageAttributes']

_Partial = Optional[tuple[int, Optional[int]]]


class LoadedMessageProvider(Protocol):
    """Generic protocol that provides access to a message's loaded contents
    when they are available.

    """

    @property
    @abstractmethod
    def loaded_msg(self) -> Optional[LoadedMessageInterface]:
        """The loaded message, if available."""


class DynamicFetchValue(FetchValue, metaclass=ABCMeta):
    """Base class for fetch values that are dynamically read from a message.

    Args:
        attribute: The fetch attribute.
        message: The message object.
        selected: The selected mailbox.

    """

    __slots__ = ['message', 'selected']

    def __init__(self, attribute: FetchAttribute, *,
                 message: MessageInterface, selected: SelectedMailbox) -> None:
        super().__init__(attribute)
        self.message: Final = message
        self.selected: Final = selected

    @abstractmethod
    def get_value(self) -> MaybeBytes:
        """Computes the value of the fetch attribute for the message."""
        ...

    def __bytes__(self) -> bytes:
        return BytesFormat(b'%b %b') % (
            self.attribute.for_response, self.get_value())


class DynamicLoadedFetchValue(FetchValue, metaclass=ABCMeta):
    """Base class for fetch values that are dynamically read from a message and
    require loading its contents.

    Args:
        attribute: The fetch attribute.
        message: The message object.
        selected: The selected mailbox.
        get_loaded: The provider of the loaded message contents, if available.

    """

    __slots__ = ['message', '_get_loaded']

    def __init__(self, attribute: FetchAttribute, *,
                 message: MessageInterface,
                 get_loaded: LoadedMessageProvider) -> None:
        super().__init__(attribute)
        self.message: Final = message
        self._get_loaded = get_loaded

    @abstractmethod
    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        """Computes the value of the fetch attribute for the message.

        Args:
            loaded_msg: The loaded message object.

        """
        ...

    def __bytes__(self) -> bytes:
        loaded_msg = self._get_loaded.loaded_msg
        if loaded_msg is None:
            value: MaybeBytes = MessageAttributes.placeholder
        else:
            value = self.get_value(loaded_msg)
        return BytesFormat(b'%b %b') % (
            self.attribute.for_response, value)

    @classmethod
    def _get_data(cls, section: FetchAttribute.Section, partial: _Partial,
                  loaded_msg: LoadedMessageInterface, *,
                  binary: bool = False) -> Writeable:
        specifier = section.specifier
        parts = section.parts
        headers = section.headers
        if specifier is None:
            data = loaded_msg.get_body(parts, binary)
        elif specifier == b'MIME' and parts is not None:
            data = loaded_msg.get_headers(parts)
        elif specifier == b'TEXT':
            data = loaded_msg.get_message_text(parts)
        elif specifier == b'HEADER':
            data = loaded_msg.get_message_headers(parts)
        elif specifier == b'HEADER.FIELDS':
            data = loaded_msg.get_message_headers(parts, headers)
        elif specifier == b'HEADER.FIELDS.NOT':
            data = loaded_msg.get_message_headers(parts, headers, True)
        else:
            raise RuntimeError()  # Should not happen.
        return cls._get_partial(data, partial)

    @classmethod
    def _get_partial(cls, data: Writeable, partial: _Partial) -> Writeable:
        if partial is None:
            return data
        full = bytes(data)
        start, end = partial
        if end is None:
            end = len(full)
        return Writeable.wrap(full[start:end])


class _UidFetchValue(DynamicFetchValue):

    def get_value(self) -> MaybeBytes:
        return Number(self.message.uid)


class _FlagsFetchValue(DynamicFetchValue):

    def get_value(self) -> MaybeBytes:
        session_flags = self.selected.session_flags
        flag_set = self.message.get_flags(session_flags)
        return List(flag_set, sort=True)


class _InternalDateFetchValue(DynamicFetchValue):

    def get_value(self) -> MaybeBytes:
        return DateTime(self.message.internal_date)


class _EmailIdFetchValue(DynamicFetchValue):

    def get_value(self) -> MaybeBytes:
        return self.message.email_id.parens


class _ThreadIdFetchValue(DynamicFetchValue):

    def get_value(self) -> MaybeBytes:
        try:
            return self.message.thread_id.parens
        except ValueError:
            return Nil()


class _LoadedMessageProvider(LoadedMessageProvider):

    __slots__ = ['loaded_msg']

    def __init__(self) -> None:
        super().__init__()
        self.loaded_msg: Optional[LoadedMessageInterface] = None

    @contextmanager
    def apply(self, loaded_msg: LoadedMessageInterface) -> Iterator[None]:
        self.loaded_msg = loaded_msg
        try:
            yield
        finally:
            self.loaded_msg = None


class _EnvelopeFetchValue(DynamicLoadedFetchValue):

    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        return loaded_msg.get_envelope_structure()


class _BodyStructureFetchValue(DynamicLoadedFetchValue):

    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        return loaded_msg.get_body_structure().extended


class _BodyFetchValue(DynamicLoadedFetchValue):

    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        attr = self.attribute
        if attr.section is None:
            return loaded_msg.get_body_structure()
        else:
            data = self._get_data(attr.section, attr.partial, loaded_msg)
            return LiteralString(data)


class _BodyPeekFetchValue(_BodyFetchValue):
    pass


class _RFC822FetchValue(_BodyFetchValue):
    pass


class _RFC822HeaderFetchValue(_BodyFetchValue):
    pass


class _RFC822TextFetchValue(_BodyFetchValue):
    pass


class _RFC822SizeFetchValue(DynamicLoadedFetchValue):

    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        return Number(loaded_msg.get_size())


class _BinaryFetchValue(DynamicLoadedFetchValue):

    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        attr = self.attribute
        if attr.section is None:
            raise RuntimeError()  # should not happen.
        data = self._get_data(attr.section, attr.partial, loaded_msg,
                              binary=True)
        return LiteralString(data, True)


class _BinaryPeekFetchValue(_BinaryFetchValue):
    pass


class _BinarySizeFetchValue(DynamicLoadedFetchValue):

    def get_value(self, loaded_msg: LoadedMessageInterface) -> MaybeBytes:
        attr = self.attribute
        if attr.section is None:
            raise RuntimeError()  # should not happen.
        data = self._get_data(attr.section, attr.partial, loaded_msg,
                              binary=True)
        return Number(len(data))


class MessageAttributes(Sequence[FetchValue]):
    """Defines the logic for how fetch attributes are resolved on a message to
    produce a fetch value.

    Args:
        message: The message object.
        selected: The selected mailbox.

    """

    #: Placeholder value for fetch values requiring loaded message contents.
    placeholder: ClassVar[bytes] = b'...'

    _simple_attrs: Mapping[bytes, type[DynamicFetchValue]] = {
        b'UID': _UidFetchValue,
        b'FLAGS': _FlagsFetchValue,
        b'INTERNALDATE': _InternalDateFetchValue,
        b'EMAILID': _EmailIdFetchValue,
        b'THREADID': _ThreadIdFetchValue}

    _loaded_attrs: Mapping[bytes, type[DynamicLoadedFetchValue]] = {
        b'ENVELOPE': _EnvelopeFetchValue,
        b'BODYSTRUCTURE': _BodyStructureFetchValue,
        b'BODY': _BodyFetchValue,
        b'BODY.PEEK': _BodyPeekFetchValue,
        b'RFC822': _RFC822FetchValue,
        b'RFC822.HEADER': _RFC822HeaderFetchValue,
        b'RFC822.TEXT': _RFC822TextFetchValue,
        b'RFC822.SIZE': _RFC822SizeFetchValue,
        b'BINARY': _BinaryFetchValue,
        b'BINARY.PEEK': _BinaryPeekFetchValue,
        b'BINARY.SIZE': _BinarySizeFetchValue}

    __slots__ = ['message', 'selected', 'attributes', 'requirement',
                 '_get_loaded', '_values']

    def __init__(self, message: MessageInterface,
                 selected: SelectedMailbox,
                 attributes: Sequence[FetchAttribute]) -> None:
        super().__init__()
        self.message: Final = message
        self.selected: Final = selected
        self.attributes: Final = attributes
        self.requirement: Final = FetchRequirement.reduce(
            attr.requirement for attr in attributes)
        self._get_loaded = _LoadedMessageProvider()
        self._values: Optional[Sequence[FetchValue]] = None

    @asynccontextmanager
    async def load_hook(self) -> AsyncIterator[None]:
        """An async context manager that loads the message content on entry,
        and releases it on exit. Fetch values that require loaded message
        content will write :attr:`.placeholder` if written outside of this
        context manager, for console or log output.

        """
        loaded_msg = await self.message.load_content(self.requirement)
        with self._get_loaded.apply(loaded_msg):
            yield

    def __iter__(self) -> Iterator[FetchValue]:
        return iter(self._get_values())

    def __getitem__(self, index):
        values = self._get_values()
        return values[index]

    def __len__(self) -> int:
        return len(self.attributes)

    def _get_values(self) -> Sequence[FetchValue]:
        if self._values is None:
            self._values = [self._get(attr) for attr in self.attributes]
        return self._values

    def _get(self, attr: FetchAttribute) -> FetchValue:
        attr_name = attr.value
        simple = self._simple_attrs.get(attr_name)
        if simple is not None:
            return simple(attr, message=self.message, selected=self.selected)
        loaded = self._loaded_attrs.get(attr_name)
        if loaded is not None:
            return loaded(attr, message=self.message,
                          get_loaded=self._get_loaded)
        raise KeyError(attr_name)




from __future__ import annotations

from datetime import datetime
from typing import Optional

from grpclib.server import Stream
from pymap import __version__ as pymap_version
from pymapadmin import __version__ as pymap_admin_version
from pymapadmin.grpc.admin_grpc import SystemBase
from pymapadmin.grpc.admin_pb2 import LoginRequest, LoginResponse, \
    PingRequest, PingResponse
from pysasl.creds import AuthenticationCredentials

from . import BaseHandler

__all__ = ['SystemHandlers']

_LoginStream = Stream[LoginRequest, LoginResponse]
_PingStream = Stream[PingRequest, PingResponse]


class SystemHandlers(SystemBase, BaseHandler):
    """The GRPC handlers, executed when an admin request is received. Each
    handler should receive a request, take action, and send the response.

    See Also:
        :class:`grpclib.server.Server`

    """

    async def Login(self, stream: _LoginStream) -> None:
        """Response to a login request. For example::

            $ pymap-admin login

        See ``pymap-admin login --help`` for more options.

        Args:
            stream (:class:`~grpclib.server.Stream`): The stream for the
                request and response.

        """
        request = await stream.recv_message()
        assert request is not None
        bearer_token: Optional[str] = None
        async with self.catch_errors('Login') as result:
            credentials = AuthenticationCredentials(
                request.authcid, request.secret, request.authzid)
            expiration: Optional[datetime] = None
            if request.HasField('token_expiration'):
                expiration = datetime.fromtimestamp(request.token_expiration)
            identity = await self.login.authenticate(credentials)
            bearer_token = await identity.new_token(expiration=expiration)
        resp = LoginResponse(result=result)
        if bearer_token is not None:
            resp.bearer_token = bearer_token
        await stream.send_message(resp)

    async def Ping(self, stream: _PingStream) -> None:
        """Respond to a ping request. For example::

            $ pymap-admin ping

        See ``pymap-admin ping --help`` for more options.

        Args:
            stream (:class:`~grpclib.server.Stream`): The stream for the
                request and response.

        """
        request = await stream.recv_message()
        assert request is not None
        resp = PingResponse(pymap_version=pymap_version,
                            pymap_admin_version=pymap_admin_version)
        await stream.send_message(resp)
        assert request is not None




from __future__ import annotations

from abc import abstractmethod
from collections.abc import Collection, Sequence
from datetime import datetime
from typing import TypeVar, Protocol

from ..bytes import Writeable
from ..flags import SessionFlags
from ..parsing.response.fetch import EnvelopeStructure, BodyStructure
from ..parsing.specials import Flag, ObjectId, FetchRequirement

__all__ = ['MessageT', 'MessageT_co', 'FlagsKey', 'CachedMessage',
           'MessageInterface', 'LoadedMessageInterface']

#: Type variable with an upper bound of :class:`MessageInterface`.
MessageT = TypeVar('MessageT', bound='MessageInterface')

#: Covariant type variable with an upper bound of :class:`MessageInterface`.
MessageT_co = TypeVar('MessageT_co', bound='MessageInterface', covariant=True)

#: Type alias for the value used as a key in set comparisons detecting flag
#: updates.
FlagsKey = tuple[int, frozenset[Flag]]


class CachedMessage(Protocol):
    """Cached message metadata used to track state changes. Used to produce
    untagged FETCH responses when a message's flags have changed, or when a
    FETCH command requests metadata of an expunged message before its untagged
    EXPUNGE response has been sent.

    This is intended to be compatible with :class:`MessageInterface`, and
    should be implemented by the same classes in most cases.

    """

    __slots__: Sequence[str] = []

    @property
    @abstractmethod
    def uid(self) -> int:
        """The message's unique identifier in the mailbox."""
        ...

    @property
    @abstractmethod
    def internal_date(self) -> datetime:
        """The message's internal date."""
        ...

    @property
    @abstractmethod
    def permanent_flags(self) -> frozenset[Flag]:
        """The permanent flags for the message."""
        ...

    @property
    @abstractmethod
    def email_id(self) -> ObjectId:
        """The message's email object ID."""
        ...

    @property
    @abstractmethod
    def thread_id(self) -> ObjectId:
        """The message's thread object ID."""
        ...

    @property
    @abstractmethod
    def flags_key(self) -> FlagsKey:
        """Hashable value that represents the current flags of this
        message, used for detecting mailbox updates.

        """
        ...

    @abstractmethod
    def get_flags(self, session_flags: SessionFlags) -> frozenset[Flag]:
        """Get the full set of permanent and session flags for the message.

        Args:
            session_flags: The current session flags.

        """
        ...


class MessageInterface(Protocol):
    """Message data such as UID, permanent flags, and when the message was
    added to the system.

    """

    __slots__: Sequence[str] = []

    @property
    @abstractmethod
    def uid(self) -> int:
        """The message's unique identifier in the mailbox."""
        ...

    @property
    @abstractmethod
    def expunged(self) -> bool:
        """True if this message has been expunged from the mailbox."""
        ...

    @property
    @abstractmethod
    def internal_date(self) -> datetime:
        """The message's internal date."""
        ...

    @property
    @abstractmethod
    def permanent_flags(self) -> frozenset[Flag]:
        """The permanent flags for the message."""
        ...

    @property
    @abstractmethod
    def email_id(self) -> ObjectId:
        """The message's email object ID, which can identify its content.

        See Also:
            `RFC 8474 5.1. <https://tools.ietf.org/html/rfc8474#section-5.1>`_

        """
        ...

    @property
    @abstractmethod
    def thread_id(self) -> ObjectId:
        """The message's thread object ID, which groups messages together.

        See Also:
            `RFC 8474 5.2. <https://tools.ietf.org/html/rfc8474#section-5.2>`_

        """
        ...

    @abstractmethod
    def get_flags(self, session_flags: SessionFlags) -> frozenset[Flag]:
        """Get the full set of permanent and session flags for the message.

        Args:
            session_flags: The current session flags.

        """
        ...

    @abstractmethod
    async def load_content(self, requirement: FetchRequirement) \
            -> LoadedMessageInterface:
        """Loads the content of the message.

        Args:
            requirement: The data required from the message content.

        """
        ...


class LoadedMessageInterface(Protocol):
    """The loaded message content, which may include the header, the body,
    both, or neither, depending on the requirements.

    It is assumed that this object contains the entire content in-memory. As
    such, when multiple :class:`MessageInterface` objects are being processed,
    only one :class:`LoadedMessageInterface` should be in scope at a time.

    """

    __slots__: Sequence[str] = []

    @property
    @abstractmethod
    def requirement(self) -> FetchRequirement:
        """The :class:`~pymap.parsing.specials.FetchRequirement` used to load
        the message content.

        """
        ...

    @abstractmethod
    def __bytes__(self) -> bytes:
        ...

    @abstractmethod
    def get_header(self, name: bytes) -> Sequence[str]:
        """Get the values of a header from the message.

        Args:
            name: The name of the header.

        """
        ...

    @abstractmethod
    def get_headers(self, section: Sequence[int]) -> Writeable:
        """Get the headers from the message part.

        The ``section`` argument indexes a nested sub-part of the message. For
        example, ``[2, 3]`` would get the 2nd sub-part of the message and then
        index it for its 3rd sub-part.

        Args:
            section: Nested list of sub-part indexes.

        """
        ...

    @abstractmethod
    def get_message_headers(self, section: Sequence[int] = None,
                            subset: Collection[bytes] = None,
                            inverse: bool = False) -> Writeable:
        """Get the headers from the message or a ``message/rfc822`` sub-part of
        the message..

        The ``section`` argument can index a nested sub-part of the message.
        For example, ``[2, 3]`` would get the 2nd sub-part of the message and
        then index it for its 3rd sub-part.

        Args:
            section: Optional nested list of sub-part indexes.
            subset: Subset of headers to get.
            inverse: If ``subset`` is given, this flag will invert it so that
                the headers *not* in ``subset`` are returned.

        """
        ...

    @abstractmethod
    def get_message_text(self, section: Sequence[int] = None) -> Writeable:
        """Get the text of the message part, not including headers.

        The ``section`` argument can index a nested sub-part of the message.
        For example, ``[2, 3]`` would get the 2nd sub-part of the message and
        then index it for its 3rd sub-part.

        Args:
            section: Optional nested list of sub-part indexes.

        """
        ...

    @abstractmethod
    def get_body(self, section: Sequence[int] = None,
                 binary: bool = False) -> Writeable:
        """Get the full body of the message part, including headers.

        The ``section`` argument can index a nested sub-part of the message.
        For example, ``[2, 3]`` would get the 2nd sub-part of the message and
        then index it for its 3rd sub-part.

        Args:
            section: Optional nested list of sub-part indexes.
            binary: True if the result has decoded any
                Content-Transfer-Encoding.

        """
        ...

    @abstractmethod
    def get_size(self, section: Sequence[int] = None) -> int:
        """Return the size of the message, in octets.

        Args:
            section: Optional nested list of sub-part indexes.

        """
        ...

    @abstractmethod
    def get_envelope_structure(self) -> EnvelopeStructure:
        """Build and return the envelope structure.

        See Also:
            `RFC 3501 2.3.5.
            <https://tools.ietf.org/html/rfc3501#section-2.3.5>`_

        """
        ...

    @abstractmethod
    def get_body_structure(self) -> BodyStructure:
        """Build and return the body structure.

        See Also:
            `RFC 3501 2.3.6
            <https://tools.ietf.org/html/rfc3501#section-2.3.6>`_

        """
        ...

    @abstractmethod
    def contains(self, value: bytes) -> bool:
        """Check the body of the message for a sub-string. This may be
        optimized to only search headers and ``text/*`` MIME parts.

        Args:
            value: The sub-string to find.

        """
        ...




from __future__ import annotations

import enum
from collections.abc import Mapping, Sequence
from typing import Optional

from pymap.bytes import BytesFormat, MaybeBytes, WriteStream, Writeable
from pymap.parsing.exceptions import NotParseable
from pymap.parsing.primitives import String, QuotedString, LiteralString

__all__ = ['Condition', 'Response', 'BadCommandResponse',
           'CapabilitiesResponse', 'GetScriptResponse', 'ListScriptsResponse']

_Capabilities = Mapping[bytes, Optional[MaybeBytes]]


class Condition(enum.Enum):
    OK = enum.auto()
    NO = enum.auto()
    BYE = enum.auto()

    def __bytes__(self) -> bytes:
        return bytes(str(self.name), 'ascii')


class Response(Writeable):

    def __init__(self, condition: Condition, *, code: MaybeBytes = None,
                 text: str = None) -> None:
        super().__init__()
        self.condition = condition
        self.code = code
        self.text = text
        self._raw: Optional[bytes] = None

    @property
    def is_bye(self) -> bool:
        return self.condition == Condition.BYE

    def write(self, writer: WriteStream) -> None:
        writer.write(bytes(self.condition))
        if self.code is not None:
            writer.write(b' (%b)' % bytes(self.code))
        if self.text is not None:
            text_bytes = self.text.rstrip('\r\n').encode('utf-8', 'replace')
            text_str = String.build(text_bytes)
            writer.write(b' ')
            text_str.write(writer)
        writer.write(b'\r\n')

    def __bytes__(self) -> bytes:
        if self._raw is None:
            self._raw = self.tobytes()
        return self._raw

    def __repr__(self) -> str:
        raw = bytes(self).decode('utf-8')
        return f'<Response {repr(raw)}>'


class BadCommandResponse(Response):

    def __init__(self, exc: NotParseable) -> None:
        super().__init__(Condition.NO, text='Bad command: %s' % exc)


class NoOpResponse(Response):

    def __init__(self, tag: Optional[bytes]) -> None:
        if tag is None:
            code: Optional[bytes] = None
        else:
            code = BytesFormat(b'TAG %b') % String.build(tag)
        super().__init__(Condition.OK, code=code)


class CapabilitiesResponse(Response):

    def __init__(self, capabilities: _Capabilities, *,
                 code: MaybeBytes = None) -> None:
        super().__init__(Condition.OK, code=code)
        self.capabilities = capabilities

    def write(self, writer: WriteStream) -> None:
        for cap_name, cap_val in self.capabilities.items():
            quoted_name = QuotedString(cap_name)
            if cap_val is None:
                line = BytesFormat(b'%b\r\n') % (quoted_name, )
            else:
                quoted_val = QuotedString(bytes(cap_val))
                line = BytesFormat(b'%b %b\r\n') % (quoted_name, quoted_val)
            writer.write(line)
        super().write(writer)


class GetScriptResponse(Response):

    def __init__(self, script_data: bytes) -> None:
        super().__init__(Condition.OK)
        self.script_data = script_data

    def write(self, writer: WriteStream) -> None:
        data_str = LiteralString(self.script_data)
        data_str.write(writer)
        writer.write(b'\r\n')
        super().write(writer)


class ListScriptsResponse(Response):

    def __init__(self, active: Optional[str], names: Sequence[str]) -> None:
        super().__init__(Condition.OK)
        self.active = active
        self.names = names

    def write(self, writer: WriteStream) -> None:
        for name in self.names:
            name_str = String.build(name)
            name_str.write(writer)
            if self.active and name == self.active:
                writer.write(b' ACTIVE\r\n')
            else:
                writer.write(b'\r\n')
        super().write(writer)




from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import Optional

from . import ResponseCode
from ..primitives import List
from ..specials import SequenceSet, ObjectId
from ...bytes import MaybeBytes, BytesFormat

__all__ = ['Capability', 'PermanentFlags', 'UidNext', 'UidValidity', 'Unseen',
           'AppendUid', 'CopyUid', 'MailboxId']


class Capability(ResponseCode):
    """Lists the capabilities the server advertises to the client.

    Args:
        server_capabilities: The list of capabilities to advertise.

    """

    def __init__(self, server_capabilities: Iterable[MaybeBytes]) -> None:
        super().__init__()
        self.capabilities = [bytes(cap) for cap in server_capabilities]
        self._raw: Optional[bytes] = None

    def __contains__(self, capability: MaybeBytes) -> bool:
        return capability in self.capabilities

    @property
    def string(self) -> bytes:
        """The capabilities string without the enclosing square brackets."""
        if self._raw is not None:
            return self._raw
        self._raw = raw = BytesFormat(b' ').join(
            [b'CAPABILITY', b'IMAP4rev1'] + self.capabilities)
        return raw

    def __bytes__(self) -> bytes:
        return BytesFormat(b'[%b]') % self.string


class PermanentFlags(ResponseCode):
    """Indicates the permanent flags available in a mailbox.

    Args:
        flags: The flags to return.

    """

    def __init__(self, flags: Iterable[MaybeBytes]) -> None:
        super().__init__()
        self.flags: Sequence[MaybeBytes] = list(flags)
        self._raw = BytesFormat(b'[PERMANENTFLAGS %b]') \
            % List(self.flags, sort=True)

    def __bytes__(self) -> bytes:
        return self._raw


class UidNext(ResponseCode):
    """Indicates the next unique identifier value of the mailbox.

    Args:
        next_: The next available message UID.

    """

    def __init__(self, next_: int) -> None:
        super().__init__()
        self._raw = b'[UIDNEXT %i]' % next_

    def __bytes__(self) -> bytes:
        return self._raw


class UidValidity(ResponseCode):
    """Indicates the mailbox unique identifier validity value.

    Args:
        validity: The UID validity value.

    """

    def __init__(self, validity: int) -> None:
        super().__init__()
        self.validity = validity
        self._raw = b'[UIDVALIDITY %i]' % validity

    def __bytes__(self) -> bytes:
        return self._raw


class Unseen(ResponseCode):
    """Indicates the message sequence ID of the first message without the
    ``\\Seen`` flag.

    Args:
        next_: The sequence ID of the message.

    """

    def __init__(self, next_: int) -> None:
        super().__init__()
        self._raw = b'[UNSEEN %i]' % next_

    def __bytes__(self):
        return self._raw


class AppendUid(ResponseCode):
    """Indicates the newly assigned UIDs and UIDVALIDITY of messages appended
    to a mailbox.

    Args:
        validity: The UID validity value.
        uids: The UIDs of the appended messages.

    See Also:
        `RFC 4315 3. <https://tools.ietf.org/html/rfc4315#section-3>`_

    """

    def __init__(self, validity: int, uids: Iterable[int]) -> None:
        super().__init__()
        self.validity = validity
        self.uids = frozenset(uids)
        self._raw = BytesFormat(b'[APPENDUID %i %b]') \
            % (validity, SequenceSet.build(self.uids))

    def __bytes__(self) -> bytes:
        return self._raw


class CopyUid(ResponseCode):
    """Indicates the original and newly assigned UIDs and UIDVALIDITY of
    messages appended to the mailbox.

    Args:
        validity: The UID validity value.
        uids: The pairs of source UID mapped to destination UID.

    See Also:
        `RFC 4315 3. <https://tools.ietf.org/html/rfc4315#section-3>`_

    """

    def __init__(self, validity: int, uids: Iterable[tuple[int, int]]) -> None:
        super().__init__()
        source_uids, dest_uids = zip(*uids)
        source_uid_set = SequenceSet.build(source_uids)
        dest_uid_set = SequenceSet.build(dest_uids)
        self._raw = b'[COPYUID %i %b %b]' \
            % (validity, bytes(source_uid_set), bytes(dest_uid_set))

    def __bytes__(self) -> bytes:
        return self._raw


class MailboxId(ResponseCode):
    """Indicates the mailbox ID, which can be used to correlate mailboxes
    between renames.

    Args:
        object_id: The mailbox ID bytestring.

    See Also:
        `RFC 8474 4.1. <https://tools.ietf.org/html/rfc8474#section-4.1>`_,
        `RFC 8474 4.2. <https://tools.ietf.org/html/rfc8474#section-4.2>`_

    """

    def __init__(self, object_id: ObjectId) -> None:
        super().__init__()
        self._raw = BytesFormat(b'[MAILBOXID %b]') % (object_id.parens, )

    def __bytes__(self) -> bytes:
        return self._raw




from __future__ import annotations

from .interfaces.backend import ServiceInterface
from .plugin import Plugin

__all__ = ['services']

#: Registers new service plugins.
services: Plugin[type[ServiceInterface]] = Plugin('pymap.service')




from __future__ import annotations

from abc import abstractmethod
from collections.abc import Collection, Iterator, Mapping, MutableMapping, \
    MutableSet, Set
from typing import TypeVar, Protocol, Any
from weakref import WeakKeyDictionary, WeakSet, WeakValueDictionary

__all__ = ['MemberInterface', 'ListenCallback', 'ClusterMetadata']

ArgT = TypeVar('ArgT')
ArgT_contra = TypeVar('ArgT_contra', contravariant=True)


class MemberInterface(Protocol):
    """A hashable type that represents a cluster member node."""

    @abstractmethod
    def __hash__(self) -> int:
        ...

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the cluster member node."""
        ...

    @property
    @abstractmethod
    def metadata(self) -> _Metadata:
        """The metadata mapping from the cluster member node."""
        ...


_Metadata = Mapping[str, bytes]
_MemberValues = WeakKeyDictionary[MemberInterface, bytes]
_MemberValuesView = Mapping[MemberInterface, bytes]


class ListenCallback(Protocol[ArgT_contra]):

    @abstractmethod
    def __call__(self, arg: ArgT_contra, /, metadata: _Metadata) -> None:
        """Called when the value associated with a local metadata key has been
        changed.

        Args:
            arg: The argument given when the callback was registered with
                :meth:`~ClusterMetadata.listen`.
            metadata: The updated local metadata mapping.

        """
        ...


class _LocalMetadata(MutableMapping[str, bytes]):

    def __init__(self) -> None:
        super().__init__()
        self._map: dict[str, bytes] = {}
        self._callbacks: WeakValueDictionary[ListenCallback[Any], object] = \
            WeakValueDictionary()

    def __getitem__(self, key: str) -> bytes:
        return self._map[key]

    def __setitem__(self, key: str, value: bytes) -> None:
        local = self._map
        curr = local.get(key)
        if curr != value:
            local[key] = value
            for callback, arg in self._callbacks.items():
                callback(arg, local)

    def __delitem__(self, key: str) -> None:
        local = self._map
        del local[key]
        for callback, arg in self._callbacks.items():
            callback(arg, local)

    def __iter__(self) -> Iterator[str]:
        return iter(self._map)

    def __len__(self) -> int:
        return len(self._map)

    def __repr__(self) -> str:  # pragma: no cover
        return repr(self._map)


class _RemoteMetadata(Mapping[str, _MemberValues]):

    def __init__(self) -> None:
        super().__init__()
        self._map: dict[str, _MemberValues] = {}

    def __getitem__(self, key: str) -> _MemberValues:
        return self._map[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._map)

    def __len__(self) -> int:
        return len(self._map)

    def _add_member(self, member: MemberInterface) -> None:
        for key, value in member.metadata.items():
            member_values = self._map.get(key)
            if member_values is None:
                self._map[key] = member_values = WeakKeyDictionary()
            member_values[member] = value

    def _del_member(self, member: MemberInterface) -> None:
        for member_values in self._map.values():
            member_values.pop(member, None)

    def __repr__(self) -> str:  # pragma: no cover
        return repr({key: dict(val) for key, val in self._map.items()})


class ClusterMetadata(MutableSet[MemberInterface]):
    """Keeps a mapping of metadata pertaining to the current instance and
    possibly other clustered instances, if the current instance is part of a
    cluster.

    The object also acts as a set of :class:`MemberInterface` objects. The
    cluster service should add new or updated members and discard them as they
    go offline.

    Args:
        init: Initial remote cluster members.

    """

    def __init__(self, init: Collection[MemberInterface] = [], /) -> None:
        super().__init__()
        self._members: WeakSet[MemberInterface] = WeakSet()
        self._local = _LocalMetadata()
        self._remote = _RemoteMetadata()
        for member in init:
            self.add(member)

    def __contains__(self, member: object) -> bool:
        return member in self._members

    def __iter__(self) -> Iterator[MemberInterface]:
        return iter(self._members)

    def __len__(self) -> int:
        return len(self._members)

    def add(self, member: MemberInterface) -> None:
        self._members.add(member)
        self._remote._add_member(member)

    def discard(self, member: MemberInterface) -> None:
        self._members.discard(member)
        self._remote._del_member(member)

    @property
    def local(self) -> MutableMapping[str, bytes]:
        """The local cluster instance metadata. Keys added, removed, and
        modified in this mapping should be disseminated to the rest of the
        cluster.

        """
        return self._local

    @property
    def remote(self) -> Mapping[str, _MemberValuesView]:
        """The remote cluster instance metadata, organized by metadata key."""
        return self._remote

    def get_all(self, key: str) -> Set[bytes]:
        """Returns the set of all known values for the metadata key, local and
        remote.

        Args:
            key: The metadata key.

        """
        results = set(self._remote[key].values())
        if key in self._local:
            results.add(self._local[key])
        return results

    def listen(self, callback: ListenCallback[ArgT], arg: ArgT) -> None:
        """Adds a callback to be run whenever the local metadata has been
        modified. Cluster services can use this to disseminate metadata
        changes.

        Note:
            The *arg* object is weakly referenced, and the callback will be
            automatically removed if it is :term:`garbage-collected <garbage
            collection>`.

        Args:
            callback: Function called with the updated local metadata mapping.
            arg: The first argument passed to *callback*.

        """
        self._local._callbacks[callback] = arg
        callback(arg, self._local)




from __future__ import annotations

import base64
import quopri
from abc import abstractmethod, ABCMeta
from email.headerregistry import ContentTransferEncodingHeader
from typing import Optional

from . import MessageHeader, MessageBody
from ..bytes import Writeable

__all__ = ['MessageDecoder']


class MessageDecoder(metaclass=ABCMeta):
    """Decodes a :class:`~pymap.mime.MessageBody` as decided by its
    ``Content-Transfer-Encoding`` header.

    Attributes:
        registry (Dict[str, :class:`MessageDecoder`]): Registry of custom
            decoders. Keys should be lower-case.

    """

    registry: dict[str, MessageDecoder] = {}

    @classmethod
    def of(cls, msg_header: MessageHeader) -> MessageDecoder:
        """Return a decoder from the message header object.

        See Also:
            :meth:`.of_cte`

        Args:
            msg_header: The message header object.

        """
        cte_hdr = msg_header.parsed.content_transfer_encoding
        return cls.of_cte(cte_hdr)

    @classmethod
    def of_cte(cls, header: Optional[ContentTransferEncodingHeader]) \
            -> MessageDecoder:
        """Return a decoder from the CTE header value.

        There is built-in support for ``7bit``, ``8bit``, ``quoted-printable``,
        and ``base64`` CTE header values. Decoders can be added or overridden
        with the :attr:`.registry` dictionary.

        Args:
            header: The CTE header value.

        """
        if header is None:
            return _NoopDecoder()
        hdr_str = str(header).lower()
        custom = cls.registry.get(hdr_str)
        if custom is not None:
            return custom
        elif hdr_str in ('7bit', '8bit'):
            return _NoopDecoder()
        elif hdr_str == 'quoted-printable':
            return _QuotedPrintableDecoder()
        elif hdr_str == 'base64':
            return _Base64Decoder()
        else:
            raise NotImplementedError(hdr_str)

    @abstractmethod
    def decode(self, body: MessageBody) -> Writeable:
        """Decode and return the decoded body of the message.

        Args:
            body: The message body.

        """
        ...


class _NoopDecoder(MessageDecoder):

    def decode(self, body: MessageBody) -> Writeable:
        return body


class _QuotedPrintableDecoder(MessageDecoder):

    def decode(self, body: MessageBody) -> Writeable:
        raw = bytes(body)
        ret = quopri.decodestring(raw)
        return Writeable.wrap(ret)


class _Base64Decoder(MessageDecoder):

    def decode(self, body: MessageBody) -> Writeable:
        raw = bytes(body)
        ret = base64.b64decode(raw)
        return Writeable.wrap(ret)
